\xedr.xpl	27-Feb-2018	Version 3.4
\Text editor based on the 6502 code by P.J.R. Boyle. This version Copyright
\ 1984-2018 by Richard Ottosen, Loren Blaney, and Larry Fish.
\This version is modified to run on the Raspberry Pi under Raspbian.
\
\This program is free software; you can redistribute it and/or modify it under
\ the terms of the GNU General Public License version 2 as published by the
\ Free Software Foundation.
\This program is distributed in the hope that it will be useful, but WITHOUT
\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
\ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
\ details.
\You should have received a copy of the GNU General Public License along with
\ this program (in the file LICENSE.TXT); if not, write to the Free Software
\ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\
\You can reach the authors at: loren.blaney@gmail.com
\
\REVISIONS:
\ NOV-84, Original version, R.O.
\ JUL-90, Modified to run on PC using 16-bit XPL, L.B.
\ AUG-90, Enhanced PC version, assembly language screen handling, L.F.
\ MAR-94, V1.9L, Modified for XPL486, large edit Buffer, cleaned up, L.B.
\ JUN-94, V1.9, Modified for 8088 world (small edit Buffer).
\ July...Sept-94, V1.9x, Modified for me to test some ideas. R.O.
\ March-95 Used "condition" for large and small versions. R.O.
\ June-95 Large buffers under Windows, Misc. bug fixes. L.F.
\ July-95 Changed Ctrl+U, N, and C to work the way they did originally.
\ Cleaned up help screens, etc. L.B.
\V2.2, DEC-10-95, Fixed occasional lockup when margining last paragraph.
\V2.3, JUN-15-96, Kluged to fix infinite loop in margin command.
\V2.3, JUN-12-2014, Fixed code in MoveCur and BackWords due to bad disk sectors.
\V3.0, 28-Jun-2015, Modified for Raspberry Pi running Raspbian Linux.
\V3.1, 03-Oct-2015, Changed order of arguments in call to GetFB intrinsic.
\V3.2, 02-Feb-2016, Ctrl+Break ($1A) aborts command line. Improve help screens.
\ 20 command lines; first 10 for semi-permanent macros.
\V3.3, 18-Jan-2017, Make compatible with changes in natrpi.s SetWind intrinsic.
\V3.4, 27-Feb-2018, Modified to work with all color depths e.g: Jessie's 32-bit,
\ and verify before discarding modified file with Esc T.
\
\NOTES:
\See Help procedure for the instructions.
\
\This text editor is always in insert mode. Characters can be rapidly inserted
\ because the text in the edit Buffer is split at the point where the cursor
\ is. Inserted characters are simply stored in the split--the rest of the
\ text does not need to be moved.
\
\If the screen looks like this:
\
\ The quick brown fox jumps over the lazy dog.
\			 ^
\			Displayed cursor
\
\The edit Buffer will look like this:
\
\ The quick brown fox jum................................ps over the lazy dog.
\		         ^				 ^
\		        CInx			        C2Inx
\
\If an "X" is inserted, the screen will look like this:
\
\ The quick brown fox jumXps over the lazy dog.
\			  ^
\			Displayed cursor
\
\ and the edit Buffer will look like this:
\
\ The quick brown fox jumX...............................ps over the lazy dog.
\		          ^				 ^
\		        CInx			        C2Inx
\
\When CInx = C2Inx, the edit Buffer is full and no more characters can be
\ inserted.

include codesr;

def	PR= 2;		\printer device number

	\Extended key codes
def	FN1KEY=59,	FN2KEY=60,	FN3KEY=61,	FN4KEY=62,
	FN5KEY=63,	FN6KEY=64,	FN7KEY=65,	FN8KEY=66,
	FN9KEY=67,	FN10KEY=68,	HOMKEY=71,	ENDKEY=79,
	PUPKEY=73,	PDNKEY=81,	UPAKEY=72,	DNAKEY=80,
	LFAKEY=75,	RTAKEY=77,	INSKEY=82,	DelKey=83,
	SHFTAB=15;

	\Extended key codes with Alt key pressed
def	AltFN1=104,	AltFN2=105,	AltFN3=106,	AltFN4=107,
	AltFN5=108,	AltFN6=109,	AltFN7=110,	AltFN8=111,
	AltFN9=112,	AltFN10=113,	AltFN11=139,	AltFN12=140;

def	AltA=30,	AltB=48,	AltC=46,	AltD=32,
	AltE=18,	AltF=33,	AltG=34,	AltH=35,
	AltI=23,	AltJ=36,	AltK=37,	AltL=38,
	AltM=50,	AltN=49,	AltO=24,	AltP=25,
	AltQ=16,	AltR=19,	AltS=31,	AltT=20,
	AltU=22,	AltV=47,	AltW=17,	AltX=45,
	AltY=21,	AltZ=44;

	\Extended key codes with Ctrl key pressed
def	CtrlFN1=94,	CtrlFN2=95,	CtrlFN3=96,	CtrlFN4=97,
	CtrlFN5=98,	CtrlFN6=99,	CtrlFN7=100,	CtrlFN8=101,
	CtrlFN9=102,	CtrlFN10=103,	CtrlFN11=137,	CtrlFN12=138;
def	CtrlPgDn=118,	CtrlEnd=117;

def	ShiftFN1=$54,	ShiftFN2=$55,	ShiftFN3=$56,	ShiftFN4=$57,
	ShiftFN5=$58,	ShiftFN6=$59,	ShiftFN7=$5A,	ShiftFN8=$5B;

def	CtrlA= $01,	CtrlB= $02,	CtrlC= $03,	CtrlD= $04,
	CtrlE= $05,	CtrlF= $06,	CtrlG= $07,	CtrlH= $08,
	CtrlI= $09,	CtrlJ= $0A,	CtrlK= $0B,	CtrlL= $0C,
	CtrlM= $0D,	CtrlN= $0E,	CtrlO= $0F,	CtrlP= $10,
	CtrlQ= $11,	CtrlR= $12,	CtrlS= $13,	CtrlT= $14,
	CtrlU= $15,	CtrlV= $16,	CtrlW= $17,	CtrlX= $18,
	CtrlY= $19,	CtrlZ= $1A,

	Bel= $07,	Tab= $09,	LF= $0A,	FF= $0C,
	CR= $0D,	EOF= $1A,	Esc= $1B,	Sp= $20,

	SpInd= ^,	\indicator for a space when showing control chars
	WildCh= CtrlW,	\wild character matches anything in a search
	PunctCh= $EF;	\punctuation character for bracketing words in a search

def	IntSize= 4,	\number of bytes in an integer
	Huge= $7FFF_FFFE,   \largest possible limit for 'for' loop
	ScoopSize= 99_999, \scoop buffer size (characters)
	Pad= 10000,	\space left when reading in a file
	RulerSize= 256,	\maximum width of tick-mark ruler (char columns)
	MarkMax=2,	\maximum number of markers
	DBufSize= 256,	\size of deferred command (line) buffer
	StatAtt= $17;	\attribute for status, ruler, help, etc. (white on blue)

int	TV,		\display device (6, $106, or $206)
	FontHeight,	\character cell height in pixels
	BufSize,	\size of edit text buffer (bytes)
	CInx,		\index into text Buffer where cursor appears
	C2Inx,		\index for following characters (see diagram)
	SCInx,		\saved char pointers for OOPS command
	SC2Inx,
	CurH,		\horizontal position of cursor from last CR
	CurL,		\line on the screen where the cursor is currently
	Width,		\width of screen (character positions)
	Height,		\total number of lines on the screen
	RealCol,	\text column shown on the status line
	LeftCol,	\column number of the left-most displayed character
	RightCol,	\column number +1 of the right-most displayed character
	HeightCur,	\line on screen where cursor appears
	StatusMode,	\flag: status lines are displayed
	StatHeight,	\number of status-type lines shown at bottom of screen
	ScrollMode,	\flag: text scrolls about fixed cursor line
	FixedCur,	\line on screen when cursor doesn't move vertically
	XHairMode,	\flag: cross hairs are displayed
	Modified,	\flag: Buffer text was modified ("Mod" in ShowStatus)
	ModDel,		\detects inserts or deletes

	DInx,		\array: index for command-line buffer (points to end +1)
	XInx,		\index for X register (scoop)
	Char,		\holds character struck on keyboard
	BaseLine,	\base line count for counting lines in a document
	CutFlag,	\flag: text cutting enabled
	ScoopOn,	\scoop is enabled
	EOFFlag,	\EOF is read from input file
	ErrFlag,	\error flag used for terminating command-line loop

	DBuf,		\array of command-line buffers
	DN;		\current DBuf (0..DNSize-1)
def	DNSize= 20;
int	CutPtr1,	\pointers for cutting text from buffer
	CutPtr2,
	CutCol1,	\column values for cutting text from buffer
	CutCol2,

	BoxChar,	\array of box graphics characters
	BoxMode,	\current box style (0..4)

	MarkPtr,	\pointer for marker array
	Marker,		\array of text markers

	CycleFlag,	\flag: cycle command lines
	BreakFlag,	\flag: Break key was struck
	WorkCol,	\working column for UpLine and DownLine (up/down arrows)
	Global;		\global variable for expression evaluation

char	Buffer,		\main text (edit) buffer
	Scoop;		\scoop buffer

char	HexStr,		\string for showing hex on command line (must be global)
	StatCond,	\status condition message string (8 chars max)
	TickMarks;	\ruler tick marks

int	Params;		\initialization parameters (can be saved with program)
def	\Params\	\note: order must agree with constant array
	LeftMargin,	\left margin column position
	RightMargin,	\right margin column position (last char is in this col)
	BellCol,	\column for end-of-line bell
	ParaIndent,	\number of spaces to indent the beginning of paragraphs
	S1Flag,		\ignore upper/lower case in searches
	S2Flag,		\search backwards
	ShowFlag,	\flag: show the control chars
	WrapFlag,	\flag: wordwrap mode
	OverFlag;	\flag: overwrite vs. insert mode

int	InHand,		\input file handle
	OutHand,	\output file handle
	HaveCR;		\input file has carriage returns (not just line feeds)
char	InFile,		\input file name
	OutFile,	\output file name
	OutTemp,	\temporary output file name (.tmp)
	OutBack;	\backup output file name (.bak)
def	NameSize= 40;	\maximum number of characters in a file name

\-------------------------------------------------------------------------------

proc	SetVideo;	\Set parameters to match video configuration
int	FB, W, H, D;
[FB:= GetFB;
W:= FB(0);
H:= FB(1);
D:= FB(2);
SetFB(W-32, H, D);	\change depth to 4 bits so colors display correctly
SetFB(W, H, 4);		\strange gyration is required to sync from 16-bit depth
Width:= W/8;
if Width > 128 then Width:= 128;
FontHeight:= 16;
if TV = $106 then FontHeight:= 8;
if TV = $206 then FontHeight:= 14;
Height:= H/FontHeight;
FixedCur:= Height-5;
HeightCur:= FixedCur;
SetWind(0, 0, 1000, 1000, 0, 0);	\disable any possible window scrolling
];	\SetVideo



proc	CopyBuffer(From, To, Size);	\Copy contents of Buffer
int	From, To, Size;
int P;
[if To > From then		\beware of overlapping Buffer segments
	for P:= Size-1 downto 0 do Buffer(To+P):= Buffer(From+P)
else	for P:= 0 to Size-1     do Buffer(To+P):= Buffer(From+P);
];	\CopyBuffer



func	LineCount;	\Return the line within the Buffer that the cursor is on
int	I, L;
[L:= 0;
for I:= 0 to CInx-1 do if Buffer(I) = CR then L:= L+1;
return L;
];	\LineCount



proc	ExtText(Dev, Str);	\Special version of RawText intrinsic
int	Dev;
char	Str;
[while Str(0) # $A0 do		\loop until termination character
	[case Str(0) of
	  ^~:	ChOut(Dev, ^^);	\(because ^ misaligns source code text)
	  ^|:	ChOut(Dev, ^~)
	other	ChOut(Dev, Str(0));
	Str:= Str+1;
	];
];	\ExtText



func	WordSep(Ch);	\Returns 'true' if character is a word separator
int	Ch;
[if Ch=Sp ! Ch=Tab ! Ch=CR then return true;
return false;
];	\WordSep



func	StrCmp(A, B);	\Compare ASCIIZ string A to B
\This returns:
\	>0 if A > B
\	=0 if A = B
\	<0 if A < B
\This provides a general string compare, for example:
\ if StrCmp(A, B) >= 0 then...	(if A >= B then...)
char	A, B;			\strings to be compared
[while A(0) = B(0) do
	[if A(0) = 0 then return 0;
	A:= A+1;  B:= B+1];
return A(0) - B(0);
];	\StrCmp

\-------------------------------------------------------------------------------

proc	ShowRuler;	\Display the ruler, e.g: /----+--@-1----+----2----+--
\Inputs: LeftCol, RightCol, TickMarks
int	I, AT, LM, RM, BL, IN;
[if KeyHit then return;		\try to keep up with more urgent display items
if not StatusMode then return;
Cursor(0, Height-3);
if RightCol < RulerSize then
	[Attrib(StatAtt);
	\Show the relative position in the edit Buffer
	AT:= CInx *Width /( (BufSize - (C2Inx-CInx)) +1) + LeftCol;
	LM:= Params(LeftMargin);
	RM:= Params(RightMargin);
	BL:= Params(BellCol);
	IN:= Params(ParaIndent) +LM;
	for I:= LeftCol to RightCol-1 do
		case I of
		 AT:	ChOut(TV, $02);					\smiley
		 BL:	ChOut(TV, if Params(WrapFlag) then $04 else ^^);\diamond
		 IN:	ChOut(TV, ^/);
		 LM,RM:	ChOut(TV, ^|)
		other ChOut(TV, TickMarks(I));
	Attrib($07);
	]
else for I:= LeftCol to RightCol-1 do ChOut(1, ^-);
];	\ShowRuler

\-------------------------------------------------------------------------------

proc	ShowCmdLine;	\Display the selected command line (DBuf, up to DInx)
\Inputs:  DInx, DBuf, DN
\Outputs: CurH, CurL
int	I, H, S, Ch;
[if KeyHit then return;
if StatHeight=0 & not StatusMode & not ErrFlag then return;
H:= 0;
Cursor(0, Height-2);
Attrib($07);				\set normal white on black attribute
S:= if DInx(DN) >= Width then DInx(DN)-Width else 0;
for I:= S to DInx(DN)-1 do
	[Ch:= DBuf(DN, I);
	if Ch < $20 then
		[Attrib($70);		\set inverse attribute
		ChOut(TV, Ch+$40);
		Attrib($07);
		]
	else	ChOut(TV, Ch);
	H:= H+1;
	];

CurH:= H;  CurL:= Height-2;		\tell KeyIn where the cursor is
					\ (kluge to help EnterCmdLine)
while H < Width do			\blank rest of command line
	[ChOut(TV, Sp);
	H:= H+1;
	];
];	\ShowCmdLine

\-------------------------------------------------------------------------------

proc	ShowStatus;	\Display status line at the bottom of the screen, e.g:
\illegal?  Cmd: 0  Scoop: 99999  S#s > ins + Rem: 123456  At: 12345   Col: 123
\WARNINGS: This routine assumes at least an 80-column screen.
\ If "Cmd:" is moved then proc SelCmd must also be changed.
\Inputs: CInx, C2Inx, CurH, DN, XInx, StatCond
int	I, F, Ch;
char	Mod;

	proc	NumOut(Num, Size);
	\Display a left-justified number in a field represented by Size
	\e.g: a 3-digit field is represented by Size = 100.
	int	Num, Size;
	[IntOut(TV, Num);
	if Num=0 then Num:= 1;
	while Size > Num do
		[ChOut(TV, Sp);
		Num:= Num*10;
		];
	];	\NumOut


[	\ShowStatus
if KeyHit then return;
if StatHeight=0 & not StatusMode & not ErrFlag then return;
Cursor(0, Height-1);
Attrib(StatAtt);

F:= true;				\show status message, make sure old
for I:= 0 to 9 do			\ message is erased, and don't flicker
	[Ch:= if F then StatCond(I) else Sp;
	if Ch>=$80 then [Ch:= Ch&$7F;  F:= false];
	ChOut(TV, Ch);
	];

Text(TV, "Cmd: ");
NumOut(DN, 1000);

Text(TV, "Scoop: ");
NumOut(ScoopSize-XInx, 100000);

Text(TV, "  S");
ChOut(TV, if Params(S1Flag) then ^= else ^#);
Text(TV, "s ");
ChOut(TV, if Params(S2Flag) then $1B else $1A);	\ <- else ->

if Params(OverFlag) then
	[Attrib($E0);			\black on yellow
	Text(TV, " OVR ");
	Attrib(StatAtt);
	]
else	Text(TV, " ins ");
ChOut(TV, if BoxMode = 0 then Sp else BoxChar(BoxMode, 5));
Text(TV, " Rem: ");
NumOut(C2Inx-CInx, 100000);

Text(TV, "  At: ");
NumOut(CInx, 100000);

Text(TV, "  Col: ");
NumOut(RealCol, 10000);

Mod:= "   Mod";
for I:= 0 to Width-81 do
	ChOut(TV, if Modified & I<=5 then Mod(I)&$7F else Sp);

string 0;
Cursor(90, Height-1);
Text(TV, InFile);
if StrCmp(InFile, OutFile) # 0 then
	[Text(TV, "   ");  Text(TV, OutFile)];
string 1;

Attrib($07);
];	\ShowStatus

\-------------------------------------------------------------------------------

proc	Display;	\Display the edit Buffer for the current cursor position
\Inputs: Buffer, BufSize, CInx, C2Inx, Width, Height, StatHeight,
\ ScrollMode, StatusMode, XHairMode, HeightCur, FixedCur, Params(ShowFlag)
\Outputs: RealCol, LeftCol, RightCol, CurH, CurL, HeightCur
int	I,		\scratch
	CLine,		\index for the start of a line
	L;		\line counter
int	C,		\scratch index into text buffer
	H,		\horz. posn. from last carriage return (as displayed)
	Ch, Ch0,	\character, and preceding character
	Att,		\attribute for syntax highlighting
	InStr,		\flag: displaying characters inside a string
	InCom,		\0=no comment; 1=toggleable comment; 2=full-line comment
	ParenNest;	\parenthetical nesting level determines syntax color


	proc	SetAtt;		\Set attribute color for syntax highlighting
	case ParenNest of
	   0:	Attrib($07);	\normal white on black
	  -1:	Attrib($03);	\cyan
	  -2:	Attrib($06);	\brown
	  -3:	Attrib($02);	\green
	  -4:	Attrib($08);	\gray
	   1:	Attrib($0D);	\bright magenta
	   2:	Attrib($0B)	\bright cyan
	other	Attrib($0E);	\yellow


	proc	DispChar(Ch);	\Display an edit-buffer character
	int	Ch;
	\Inputs and outputs: H, L


		proc	DispX(Ch);
		int	Ch;
		[if H>=LeftCol & H<RightCol then	\character is visible
		    [if XHairMode then			\do syntax highlighting
			[if Ch0=^^ & InCom=0 then	\don't highlight ^(...
			    [ChOut(TV, Ch);  Ch0:= 0]	\but do highlight ^^)...

			else if Ch0 = ^/ then		\C comments /* and //
			  [if Ch=^* ! Ch=^/ then
			    if InCom=0 & ~InStr then
			      [Attrib($09);  InCom:= 2];
			  ChOut(TV, Ch);  Ch0:= Ch]

			else [case Ch of
			    ^(:	[ChOut(TV, Ch);		\colorize parenthetical
				if ~InStr & InCom=0 then \ text based on
				    [ParenNest:= ParenNest-1;   \ nesting level
				    SetAtt]];
			    ^):	[if ~InStr & InCom=0 then
				    [ParenNest:= ParenNest+1;
				    SetAtt];
				ChOut(TV, Ch);
				];
			    ^":	if InStr then		\strings are light red
					[ChOut(TV, Ch);
					if InCom = 0 then
					    [SetAtt;  InStr:= false]]
				else	[if InCom = 0 then
					    [Attrib($0C);  InStr:= true];
					ChOut(TV, Ch)];
			    ^\:	[if Ch0 = ^\ then InCom:= 2;
				if InCom # 0 then	\comments are light blue
					[ChOut(TV, Ch);
					if ~InStr then InCom:= InCom & ~1;
					if InCom = 0 then SetAtt]
				else	[if ~InStr then
					    [Attrib($09);  InCom:= InCom ! 1];
					ChOut(TV, Ch)];
				]
			     other	ChOut(TV, Ch);
			     Ch0:= Ch;
			     ];
			]
		    else ChOut(TV, Ch);
		    ];
		H:= H+1;				\bump horz. position
		];	\DispX


	[	\DispChar
	if Ch >= $21 then DispX(Ch)
	else	[if Params(ShowFlag) then	\show normally hidden Ctrl chars
			[if Ch = Sp then DispX(SpInd)
			else	[Attrib($70);	 \inverse video
				DispX(Ch ! $40); \convert to a character
				Attrib($07);	 \restore normal video
				];
			]
		else	[if Ch = Sp then DispX(Sp)
			else if Ch = Tab then
				repeat DispX(Sp) until (H & $07) = 0
			else if Ch = CR then []
			else	[Attrib($70);	 \inverse video
				DispX(Ch ! $40); \convert to a character
				Attrib($07);	 \restore normal video
				];
			];

		if Ch = CR then			\blank rest of the line
			[InStr:= false;
			InCom:= 0;
			ParenNest:= 0;
			SetAtt;
			while H < RightCol do DispX(Sp);
			CrLf(TV);		\don't assume screen wrap-around
			L:= L+1;		\new line
			H:= 0;			\left column
			];
		];
	];	\DispChar


[	\Display
InStr:= false;			\turn off any syntax highlighting
InCom:= 0;
ParenNest:= 0;
Ch0:= 0;
SetAtt;

\Determine how much horizontal scroll (LeftCol) is needed (this is slightly
\ complicated because of tabs):
C:= CInx;				\move C to the start of the line
while Buffer(C-1)#CR & C>0 do C:= C-1;	\ that the cursor is on
CLine:= C;				\save the start of this line

\Advance to just before the cursor, counting character positions along the way
H:= 0;					\horizontal character position
for I:= CLine to CInx-1 do
	[H:= H+1;
	if Buffer(I)=Tab & not Params(ShowFlag) then
		H:= H+7 & ~7;		\move to next tab stop
	];
LeftCol:= H+1+3 - Width;		\show at least 3 chars after cursor
if LeftCol < 0 then LeftCol:= 0;
RightCol:= Width + LeftCol;		\right-most displayed column +1
RealCol:= H;
CurH:= H;  CurL:= HeightCur-1;		\save cursor position

\Move to the start of the first displayed line or to the start of the Buffer
L:= HeightCur-1;			\maximum number of lines to back up
while CLine>0 & L>0 do
	[if CLine > 0 then CLine:= CLine-1;	\back up over the CR
	while Buffer(CLine-1)#CR & CLine>0 do	\back up a line
		CLine:= CLine-1;
	L:= L-1;
	];
CurL:= CurL-L;

\Display characters up to the cursor position
Cursor(0, 0);
H:= 0;
C:= CLine;
while C < CInx & not KeyHit do
	[DispChar(Buffer(C));
	C:= C+1;
	];

\Display the rest of the text from the cursor down to the bottom of the screen
C:= C2Inx;
L:= CurL;
while C<BufSize & L<Height-StatHeight & not KeyHit do
	[DispChar(Buffer(C));  C:= C+1];

\If at the end of the buffer, fill out the rest of the display area with spaces
while  L < Height-StatHeight do
	[while H < RightCol & not KeyHit do	\fill out line with spaces
		[if H >=LeftCol then ChOut(TV, Sp);
		H:= H+1;	\bump horz. position
		];
	CrLf(TV);		\don't assume screen wrap-around
	L:= L+1;		\new line
	H:= 0;
	];

\Display the cursor as a bright red highlighted character
Cursor(if CurH<Width-3 then CurH else Width-1-3, CurL);
Attrib($C0);
if C2Inx < BufSize then
	[Ch:= Buffer(C2Inx);
	if Ch >= $21 then ChOut(TV, Ch)
	else	if Params(ShowFlag) then
		      ChOut(TV, if Ch=Sp then SpInd else Ch+$40)
		else  ChOut(TV, if Ch=Sp ! Ch=Tab ! Ch=CR then Sp else Ch+$40);
	]
else	ChOut(TV, Sp);
Attrib($07);

if XHairMode then
	[H:= (CurL+1)*FontHeight-1;			\cross hairs
	Move(0, H);  Line(Width*8-1, H, $5500_001F);	\dotted blue line
	H:= if CurH<Width-3 then CurH else Width-1-3;
	Move(H*8+7, 0);
	Line(H*8+7, (Height-StatHeight)*FontHeight-1, $5500_001F);
	];

if CurH = Params(BellCol) then ChOut(0, Bel);

ShowRuler;
ShowStatus;
];	\Display

\-------------------------------------------------------------------------------

proc	SetStat(Str, Flag);	\Set status condition
char	Str, Flag;		\Str can have 8 characters maximum
\Inputs:  ErrFlag
\Outputs: StatCond, ErrFlag
[if not ErrFlag ! not Flag then	\don't set another condition if we already have
	[StatCond:= Str;	\ an error condition--only clear the error
	ErrFlag:= Flag;
	];
if ErrFlag then
	[if StatusMode then StatHeight:= 3
	else [StatHeight:= 2;  ShowStatus;  ShowCmdLine];
	]
else if StatusMode then StatHeight:= 3 else StatHeight:= 0;
];	\SetStat



proc	StatusOnOff;	\Turn on and off the display of status, ruler and
[if StatusMode then	\ command lines
	[StatusMode:= false;
	StatHeight:= 0;
	Display;
	]
else	[StatusMode:= true;
	StatHeight:= 3;
	HeightCur:= FixedCur;
	Display;
	ShowCmdLine;
	];
];	\StatusOnOff



proc	XHairOnOff;	\Turn on and off the display of the cross hairs
[if XHairMode then
	XHairMode:= false
else	XHairMode:= true;
Display;
];	\XHairOnOff



proc	ScrollOnOff;	\Turn on and off scrolling of the display
[if ScrollMode then
	[ScrollMode:= false;
	HeightCur:= FixedCur;
	Display;
	]
else	[ScrollMode:= true;
	HeightCur:= FixedCur;
	Display;
	];
];	\ScrollOnOff

\-------------------------------------------------------------------------------

func	KeyInX;		\Get char from keyboard (lowest level)
int	Ext, Ch;
[Ch:= GetKey;

if BoxMode # 0 then
\\    if (Ext>=HOMKEY & Ext<=INSKEY) then	\only use numeric keypad
	case of			\translate box mode characters
	  Ch>=^0 & Ch<=^9:	Ch:= BoxChar(BoxMode, Ch-^0);
	  Ch=^+:		Ch:= BoxChar(BoxMode, 10)
	other [];
return Ch;
];	\KeyInX



func	KeyIn(Att);	\Get a character from the keyboard, with cursor
int	Att;
\Inputs: CurH, CurL, LeftCol
int	Ch, H;
[H:= CurH - LeftCol;
if H<0 then H:= 0;
if H > Width-1 then H:= Width-1;
Cursor(H, CurL);

Attrib(Att);  ChOut(TV,^);		\display cursor
Ch:= KeyInX;				\get the character

Cursor(H, CurL);
return Ch;
];	\KeyIn

\-------------------------------------------------------------------------------

proc	Help(Dev);	\Show (or print) help screens
int	Dev;		\output device
int	Page, P;
def	PMax=4;		\maximum page number (= number of screens)
[Page:= [
"    XED 3.4                 -- DIRECT COMMANDS --
ͻ
       MOVE                                 ADDITIONAL CHARACTERS           
 ~A    Ahead a character              ~I    Tab                             
 ~Q    Back a character               ~L    Form feed (new page)            
 ~N |N Ahead to Next word             ~U    Indent same as line above       
 ~B |B Back a word                    ~Z    Shift case and move ahead       
 ~S    Down a line                    ~T    Display control chars (on/off)  
 ~W    Up a line                      ~~    Always insert next keystroke    
 ~J    Jump ahead a page (PgDn)      Ķ
 ~K    Kick back a page (PgUp)              SCOOP                           
 ~D    Ahead to end of edit buffer    ~R    Start (or Resume) a scoop-up    
 ~E    Back to start of edit buffer   ~F    Finish a scoop-up               
                                      ~Y    Insert scoop and empty it       
Ķ
       DELETE                               COMMAND LINE                    
 Bksp  Back a character (~H)          Esc   Enter command-line mode         
 Del   Ahead a character              ~G    Re-execute command line (Go)    
 ~C |C Back a word                    ~V    Copy command line into text     
 ~X    Back a line                    ~P    Select command line (0-9)       
 |X    Ahead a line                   ~]    Change search case sensitivity  
                                      ~_    Reverse search direction        
ͼ
        ~ = Ctrl key    | = Alt key    Fn = Function key    SFn = Shifted ",

"                            -- DIRECT COMMANDS --
  ͻ
   F1    Get last command line        SF1    Display help summary         
   F2    Get next command line        SF2    Display ASCII chart          
   F3    Turn wordwrap mode on/off    SF3    Show status lines (off/on)   
   F4    Get next buffer from file    SF4    Highlight syntax (on/off)    
   F5    Insert scoop                 SF5    Insert column scoop          
   F6    Margin paragraph                                                 
   F7    Display current line number  SF7    Scroll text (on/off)         
   F8    Erase scoop                  SF8    Select font size (S/M/L)     
   F9    Mark location in text                                            
   F10   Move to last marker                                              
  Ķ
                                      |R     Start column scoop-up        
                                      |F     Finish column scoop & blank  
                                      |Y     Insert column scoop & empty  
  Ķ
   ~O |O   Undelete (Oops!)                                                
   Insert  Select insert or overwrite mode                                 
   |D      Draw with box characters:                                   
   ~Break  Break out of a command-line loop                                
                                                                           
  ͼ
       ~ = Ctrl key    | = Alt key    Fn = Function key    SFn = Shifted ",

"                         -- COMMAND-LINE COMMANDS --
 ͻ
       DELETE                              QUIT                            
  D    Back a character               T    Terminate and discard text      
  R    Rest of line, including CR     Q    Quit and save text              
  E    Erase scoop                                                         
  Z    Zap back to form feed (~L)    Ķ
  K    Kill entire edit buffer             MISCELLANY                      
 Ĵ /    List help summary on printer    
       INSERT                         L    List edit buffer on printer     
  I    String                         M    Set left, right margins or bell 
  X    Scoop                          Y    Center line                     
  H    Character by its Hex value     V    Copy text into command line     
                                      J    Re-execute command line (Jump)  
 Ĵ P    Execute command line 0-9        
       FILE                           Esc  Command separator               
  G    Get next buffer from file      Esc Esc  Execute command line        
  A    Append from input file         { }  Enclose a loop                  
  W    Write to output file           ;    Exit a loop                     
  O    Open input and output files    #    Repeat command indefinitely     
  C    Close output file                                                   
                                                                           
 ͼ
 ",

"                         -- COMMAND-LINE COMMANDS --
 ͻ
       EXPRESSION (sets variable)          THE VARIABLE                    
  ( )  Enclose an expression          BA   Insert it as an ASCII character 
  [ ]  Execute if variable is not 0   BH   Insert it as a Hex number       
  $    Hex constant                   BD   Insert it as a Decimal number   
  ~    ASCII constant                 BN   Show on status line in decimal  
  %    The variable                   BS   Show on Status line in hex      
  @    Character under cursor         BT   Set it to scooped decimal value 
  K    Keystroke (waits)              BX   Set it to scooped hex value     
  Z    Space remaining in buffer      :    Move to location in variable    
  .    Cursor location in buffer      _    Set loop count with variable    
  C    Cursor column position        Ķ
  + -  Add, Subtract                       SEARCH                          
  * /  Multiply, Divide               S    For a string                    
  \    Remainder                             ~W = Wild character           
  & !  And, Or                               |W = Word separator           
  = #  Equal, Not equal               N    Entire file (N edit buffers)    
  > <  Greater than, Less than        F    Find and replace                
  >=   Greater than or equal          + -  Search forward or backward      
  <=   Less than or equal                                                  
 ͼ
        ~ = Ctrl key    | = Alt key    Fn = Function key    SFn = Shifted
 "];

if Dev = PR then
	[OpenO(Dev);
	ExtText(Dev, Page(0));
	CrLf(Dev);  CrLf(Dev);  CrLf(Dev);
	ExtText(Dev, Page(1));
	CrLf(Dev);  ChOut(Dev, FF);
	ExtText(Dev, Page(2));
	CrLf(Dev);  CrLf(Dev);  CrLf(Dev);
	ExtText(Dev, Page(3));
	CrLf(Dev);  ChOut(Dev, FF);
	Close(Dev);
	]
else	[P:= 0;
	Attrib($07);
	loop	[ChOut(0, FF);
		ExtText(Dev, Page(P));
		Cursor(72, 0);  ChOut(Dev, P+^1);	\page no.
		ChOut(Dev, ^/);  ChOut(Dev, PMax+^0);
		Hilight(0, 0, 78, 0, StatAtt);

		Cursor(9, Height-1);
		Text(TV,
		"Press PgDn / PgUp for more help, or Esc to continue editing.");
		Hilight(0, Height-1, 79, Height-1, StatAtt);
		CurL:= Height-1;  CurH:= 71;	\set cursor position for KeyIn

		case KeyIn($0C) of
		  -PUPKEY, -UPAKEY: P:= P-1;
		  Esc:	   quit
		other P:= P+1;
		Attrib($07);
		if P >= PMax then P:= 0;
		if P < 0 then P:= PMax-1;
		];
	Attrib($07);
	];
];	\Help

\-------------------------------------------------------------------------------

proc	Hex2Out(Dev, N);	\Output two hex digits (a byte)
int	Dev, N;
char	HexDigit;
[HexDigit:= "0123456789ABCDEF ";
ChOut(Dev, HexDigit(N>>4 & $0F));
ChOut(Dev, HexDigit(N & $0F));
];	\Hex2Out



func	IsHexDigit(Ch);		\Is character a hexadecimal digit?
int	Ch;
case of
  Ch>=^0 & Ch<=^9:  return true;
  Ch>=^A & Ch<=^F:  return true;
  Ch>=^a & Ch<=^f:  return true
other return false;



func	AToHex(Ch);		\Convert ASCII hex digit to binary
int	Ch;
case of
  Ch>=^0 & Ch<=^9:  return Ch-^0;
  Ch>=^A & Ch<=^F:  return Ch-^A+10;
  Ch>=^a & Ch<=^f:  return Ch-^a+10
other return 0;



func	DoASCII;		\Show chart of all the ASCII characters
int	H, L, CurH0, C;
int	X1, Y1, X2, Y2;
def	X0=13, Y0=3;		\coordinates of first character in chart
[CurH0:= CurH;			\(because of wordwrap)
ChOut(0, FF);
Attrib($07);
Cursor(30, 0);
Text(TV, "-- ASCII CHART --");
Hilight(0, 0, 78, 0, StatAtt);

for L:= 0 to $F do		\label vertical axes
	[Cursor(X0-4, Y0+L+L/4);
	Hex2Out(TV, L);
	Cursor(X0+54, Y0+L+L/4);
	Hex2Out(TV, L);
	];

for H:= 0 to $F do
	[Cursor(X0-1+H*3+H/4*2, Y0-1);
	Hex2Out(TV, H<<4);
	Cursor(X0-1+H*3+H/4*2, Y0+19);
	Hex2Out(TV, H<<4);
	for L:= 0 to $F do
		[Cursor(X0+H*3+H/4*2, Y0+L+L/4);
		ChOut(TV, H<<4 ! L);
		];
	];

X1:= X0 -4;	Y1:= Y0-1;
X2:= X0 +54;	Y2:= Y0+19;
Hilight(X1, Y1,   X2+1, Y1, StatAtt);
Hilight(X1, Y2,   X2+1, Y2, StatAtt);
Hilight(X1, Y1+1, X1+1, Y2-1, StatAtt);
Hilight(X2, Y1+1, X2+1, Y2-1, StatAtt);

Cursor(9, Height-1);
Text(TV, "Enter hex code for ASCII character, or Esc to continue editing: ");
Hilight(0, Height-1, 79, Height-1, StatAtt);
CurL:= Height-1;  CurH:= 73;	\set cursor position for KeyIn

C:= KeyIn($0C);
if IsHexDigit(C) then
	[Attrib(StatAtt);
	ChOut(TV, C);		\echo keystroke
	CurH:= 74;		\set cursor position for KeyIn
	H:= C;
	C:= KeyIn($0C);
	if IsHexDigit(C) then C:= AToHex(H)<<4 + AToHex(C)
	else C:= 0;
	]
else	C:= 0;
CurH:= CurH0;		\restore to original value (because of wordwrap)
Attrib($07);
return C;
];	\DoASCII

\=============================== BASIC COMMANDS ================================

proc	SetCurLine(Dist);	\Set the cursor position vertically on screen
int	Dist;
int	H;
[if ScrollMode then return;
if C2Inx >= BufSize-1 then return;
H:= HeightCur + Dist;
if H>=1 & H<=Height-StatHeight then HeightCur:= H;
if CInx = 0 then HeightCur:= 2;
];	\SetCurLine



proc	GetColumn;	\Calculate current column number
\Display normally does this, but during macros, display is not updated
int	C, H, CLine, I;
[C:= CInx;
while Buffer(C-1)#CR & C>0 do C:= C-1;
CLine:= C;
H:= 0;
for I:= CLine to CInx-1 do
	[if Buffer(I) = Tab then if not Params(ShowFlag) then H:= H ! 7;
	H:= H+1;
	];
RealCol:= H;
];	\GetColumn



proc	SavePtrs;	\Save edit-buffer pointers in case of oops (undo).
\Deleted characters reside in the gap, and they can be restored simply by
\ restoring the edit pointers. These pointers are saved whenever a command,
\ such as a cursor move, messes up the gap.
[SCInx:= CInx;
SC2Inx:= C2Inx;
];	\SavePtrs



proc	Oops;		\Undo previous deletes
[\Restore pointers back to last saved position
if CInx # SCInx ! C2Inx#SC2Inx then [CInx:= SCInx;  C2Inx:= SC2Inx]
else SetStat("nothing?", true);
];	\Oops



proc	AddMargin(P);	\Add a margin at point P
int	P;
int	Spaces;
[Spaces:= Params(LeftMargin);
if CInx+Spaces >= C2Inx then [SetStat("full", true);  return];
CopyBuffer(P, P+Spaces, CInx-P);
CInx:= CInx + Spaces;
for P:= P to P+Spaces-1 do Buffer(P):= Sp;
];	\AddMargin



proc	WrapLine;	\Wrap end of current line onto next line
int	P, C;
[P:= CInx;
for C:= 1 to 16 do
	[if P <= 0 then return;
	P:= P-1;
	if Buffer(P) = Sp then
		[Buffer(P):= CR;
		AddMargin(P+1);
		return;
		];
	];
];	\WrapLine



proc	Insert(Ch);	\Insert a char into the edit Buffer
int	Ch;
[if CInx < C2Inx then			\if there's room...
	[Buffer(CInx):= Ch;
	CInx:= CInx+1;
	if Params(OverFlag) then
		if C2Inx<BufSize & Buffer(C2Inx)#CR & Ch#CR then
			C2Inx:= C2Inx+1;
	if Params(WrapFlag) then
		if CurH>Params(RightMargin) then [if Ch#CR then WrapLine]
		else if Ch=CR then AddMargin(CInx);
	SavePtrs;
	Modified:= true;		\also detects overwrites
	]
else	SetStat("full?", true);
];	\Insert

\-------------------------------------------------------------------------------

func	MoveCur(Loc);	\Move to location Loc within text buffer
int	Loc;
int	Size;
[if Loc > BufSize - C2Inx + CInx then return false;
if Loc = CInx then return true;
if Loc < CInx then
	[Size:= CInx - Loc;
	CopyBuffer(Loc, C2Inx-Size, Size);
	C2Inx:= C2Inx - Size;
	CInx:= Loc;
	]
else	[Size:= Loc -CInx;
	CopyBuffer(C2Inx, CInx, Size);
	CInx:= CInx + Size;
	C2Inx:= C2Inx + Size;
	];
return true;
];	\MoveCur



proc	DelChar;	\Delete back one character
[if CInx > 0 then
	[CInx:= CInx-1;
	if ScoopOn then			\scoop-up buffer is on
		[if XInx < ScoopSize then
			[Scoop(XInx):= Buffer(CInx);
			XInx:= XInx+1;
			]
		else	[SetStat("full?", true);
			CInx:= CInx+1;	\delete failed, undo it
			];
		];
	]
else	SetStat("start?", true);
];	\DelChar



proc	Delete;		\Delete or blank, depending on overwrite flag
[DelChar;
if Params(OverFlag) then	\if overwrite mode then blank with a space
    if not ErrFlag & Buffer(CInx)#CR then
	[C2Inx:= C2Inx-1;  Buffer(C2Inx):= Sp;
	Modified:= true;
	];
];	\Delete



proc	DelForward;	\Delete forward one character
[if C2Inx < BufSize then C2Inx:= C2Inx+1
else SetStat("end?", true);
];	\DelForward



proc	OldDelWord;	\Delete back one word
int	Ch;
[repeat	Delete;				\delete all word separators
	Ch:= Buffer(CInx);
until	not WordSep(Ch) ! ErrFlag;

loop	[Ch:= Buffer(CInx-1);		\delete the word
	if WordSep(Ch) ! ErrFlag then quit;
	Delete;
	if CInx = 0 then quit;
	];
];	\OldDelWord



proc	DelWord;	\Delete back one word
int	Ch;
[loop	[Ch:= Buffer(CInx-1);		\back over the word
	if Ch=CR ! Ch=Sp ! Ch=Tab ! ErrFlag then quit;
	DelChar;
	if CInx = 0 then quit;
	];
loop	[Ch:= Buffer(CInx-1);		\back over all word separators
	if Ch#CR & Ch#Sp & Ch#Tab ! ErrFlag then quit
	else if Ch=CR & Buffer(CInx-2)=CR then [DelChar;  quit];
	DelChar;
	if CInx = 0 then quit;
	];
];	\DelWord



proc	DelFwdWord;	\Delete forward one word
\Note: This routine gives an error when moving to an EOF (not just beyond it)
int	Ch;
[repeat	DelForward;			\delete forward to non-word separator
	Ch:= Buffer(C2Inx);
until	Ch#CR & Ch#Sp & Ch#Tab ! ErrFlag;

if C2Inx < BufSize then
	[loop	[Ch:= Buffer(C2Inx);	\delete forward to word separator
		if Ch=CR ! Ch=Sp ! Ch=Tab ! ErrFlag then quit;
		DelForward;
		];
	]
else	SetStat("end?", true);
];	\DelFwdWord



proc	DelLine;	\Delete back one line
[repeat DelChar until Buffer(CInx-1)=CR ! CInx=0 ! ErrFlag;
];	\DelLine			(ErrFlag? because scoop might be full)



proc	DelAll;		\Delete the entire edit Buffer (kill)
[CInx:= 0;		\Notice that the scoop is not used here
C2Inx:= BufSize;
];	\DelAll



proc	Zap;		\Delete back to previous form feed or start
			\ (the form feed is not deleted)
[repeat DelChar until Buffer(CInx-1)=FF ! CInx=0 ! ErrFlag;
];	\Zap

\-------------------------------------------------------------------------------

proc	BackChar;	\Move back one character in Buffer
[DelChar;				\move CInx back, checking for scoop, etc
if not ErrFlag then			\if the delete succeeded then...
	[C2Inx:= C2Inx-1;
	Buffer(C2Inx):= Buffer(CInx);
	SavePtrs;
	if Buffer(CInx) = CR then SetCurLine(-1);
	];
];	\BackChar



proc	OldBackWord;	\Move to the start of the current word or previous word
int	Ch;
[repeat	BackChar;			\back over all word separators
	Ch:= Buffer(CInx);
until	not WordSep(Ch) ! ErrFlag;

loop	[Ch:= Buffer(CInx-1);		\back over the word
	if WordSep(Ch) ! ErrFlag then quit;
	BackChar;
	if CInx = 0 then quit;
	];
];	\OldBackWord



proc	BackWord;   \Move to the start of the current word or the previous word
int	Ch;
[loop	[Ch:= Buffer(CInx-1);		\back over the word
	if Ch=CR ! Ch=Sp ! Ch=Tab ! ErrFlag then quit;
	BackChar;
	if CInx = 0 then quit;
	];

loop	[Ch:= Buffer(CInx-1);		\back over all word separators
	if Ch#CR & Ch#Sp & Ch#Tab ! ErrFlag then quit;
	BackChar;
	if CInx = 0 then quit;
	];
];	\BackWord



proc	BackLX;  \Standard back up a line, handling scoop problems with no limit
[repeat BackChar until Buffer(CInx-1)=CR ! CInx=0 ! ErrFlag;
];	\BackLX



proc	BackLine;	\Move back to start of the current line or previous line
int	Limit, Size, P, C;
[if ScoopOn then BackLX
else	[if CInx = 0 then [SetStat("start?", true);  return];
	Limit:= if CInx > 256 then 255 else CInx-1;
	P:= CInx-1;
	loop	[for C:= 1 to Limit do
			[if Buffer(P-1) = CR then quit;
			P:= P-1;
			];
		quit;
		];
	if Buffer(CInx-1) = CR then SetCurLine(-1);
	Size:= CInx-P;
	CopyBuffer(P, C2Inx-Size, Size);
	C2Inx:= C2Inx-Size;
	CInx:= P;
	SavePtrs;
	];
];	\BackLine



proc	BackPage;	\Move backward one page
int	I, Tmp;
[\if not ScrollMode then HeightCur:= Height - StatHeight - 1;
Tmp:= HeightCur;
if Buffer(CInx-1) # CR then BackLine;
if not ScrollMode then for I:= 1 to Height-StatHeight-1 do BackLine
else for I:= 1 to Height-StatHeight do BackLine;
HeightCur:= Tmp;
];	\BackPage



proc	BackAll;	\Move to the beginning of buffer
[if ScoopOn then repeat BackChar until CInx=0 ! ErrFlag
else	[CopyBuffer(0, C2Inx-CInx, CInx);
	C2Inx:= C2Inx - CInx;
	CInx:= 0;
	SavePtrs;
	];
if not ScrollMode then HeightCur:= 1;
];	\BackAll

\-------------------------------------------------------------------------------

proc	FwdChar;	\Move forward one char in buffer
[if C2Inx < BufSize then
	[Buffer(CInx):= Buffer(C2Inx);
	CInx:= CInx+1;  C2Inx:= C2Inx+1;
	if ScoopOn then
		if XInx > 0 then XInx:= XInx-1
		else ScoopOn:= false;
	SavePtrs;
	if Buffer(CInx-1) = CR then SetCurLine(1);
	]
else	SetStat("end?", true);
];	\FwdChar



proc	OldFwdWord;	\Move to the start of the next word
\Note: This routine gives an error when moving to an EOF (not just beyond it)
int	Ch;
[if C2Inx < BufSize then
	[loop	[Ch:= Buffer(C2Inx);	\move forward to word separator
		if WordSep(Ch) ! ErrFlag then quit;
		FwdChar;
		];
	]
else	SetStat("end?", true);

repeat	FwdChar;			\move forward to non-word separator
	Ch:= Buffer(C2Inx);
until	not WordSep(Ch) ! ErrFlag;
];	\OldFwdWord



proc	FwdWord;	\Move to the start of the next word
\Note: This routine gives an error when moving to an EOF (not just beyond it)
int	Ch;
[repeat	FwdChar;			\move forward to non-word separator
	Ch:= Buffer(C2Inx);
until	Ch#CR & Ch#Sp & Ch#Tab ! ErrFlag;

if C2Inx < BufSize then
	[loop	[Ch:= Buffer(C2Inx);	\move forward to word separator
		if Ch=CR ! Ch=Sp ! Ch=Tab ! ErrFlag then quit;
		FwdChar;
		];
	]
else	SetStat("end?", true);
];	\FwdWord



proc	FwdLX;		\Do standard line forward, handling scoop and no limit
[repeat FwdChar until Buffer(CInx-1)=CR ! ErrFlag;
];	\FwdLX



proc	FwdLine;	\Move to the start of the next line
int	Limit, P, Size;
[if ScoopOn then FwdLX
else	[if C2Inx >= BufSize then [SetStat("end?", true);  return];
	Limit:= if (BufSize-C2Inx) > 256 then C2Inx+255 else BufSize-1;
	loop	[for P:= C2Inx to Limit do
			if Buffer(P) = CR then [SetCurLine(1);  quit];
		P:= P-1;
		quit;
		];
	Size:= P+1 - C2Inx;
	CopyBuffer(C2Inx, CInx, Size);
	C2Inx:= P+1;
	CInx:= CInx+Size;
	SavePtrs;
	];
];	\FwdLine



proc	FwdPage;	\Move forward one page
int	I, Tmp;
[Tmp:= HeightCur;
if not ScrollMode then
	[HeightCur:= 1;
	for I:= 1 to Height-StatHeight-1 do FwdLine;
	]
else	for I:= 1 to Height-StatHeight do FwdLine;
HeightCur:= Tmp;
];	\FwdPage



proc	FwdAll;		\Move to the end of buffer
[if ScoopOn then	\if the scoop was on, it's gone now
	[XInx:= 0;  ScoopOn:= false];
CopyBuffer(C2Inx, CInx, BufSize-C2Inx);
CInx:= CInx + BufSize - C2Inx;
C2Inx:= BufSize;
SavePtrs;
if not ScrollMode then HeightCur:= Height - StatHeight - 1;
];	\FwdAll

\-------------------------------------------------------------------------------

func	ColumnMove(WC);	\Move cursor forward to column WC
int	WC;
int	C;
[for C:= 0 to WC-1 do
	[if ErrFlag then return C;
	case Buffer(C2Inx) of
	  Tab:	C:= C ! 7;
	  CR:	return C+1
	other	[];
	FwdChar;
	];
return C+1;
];	\ColumnMove



proc	UpLine;		\Move back one line to same column
[if CInx = 0 then [SetStat("start", true);  return];
if CurH > 256 then [BackLine;  return];
if Buffer(CInx-1) # CR then BackLine;
BackLine;
if ErrFlag then return;
if WorkCol < 0 then WorkCol:= RealCol;
ColumnMove(WorkCol);
];	\UpLine



proc	DownLine;	\Move forward one line to same column
[FwdLine;
if WorkCol < 0 then WorkCol:= RealCol;
if ErrFlag then return 1;
if CurH > 256 then return WorkCol+1;
return ColumnMove(WorkCol);
];	\DownLine



proc	ShowLineNo;	\Display the line number the cursor is on
int	L, P;
[L:= LineCount(Buffer, CInx) + BaseLine;

OpenO(8);  OpenI(8);
IntOut(8, L+1);
ChOut(8, EOF);		\for safety
P:= -1;
repeat P:= P+1;  HexStr(P):= ChIn(8) until HexStr(P) = EOF;
if P > 0 then HexStr(P-1):= HexStr(P-1) ! $80 else HexStr(0):= $A0;
SetStat(HexStr, false);
];	\ShowLineNo

\-------------------------------------------------------------------------------

proc	DelRest;	\Delete the rest of a line
[loop	[if C2Inx < BufSize then C2Inx:= C2Inx+1
	else [SetStat("end?", true);  quit];
	if Buffer(C2Inx-1) = CR then quit;
	];
];	\DelRest



proc	FlipCase;	\Flip case of letter at cursor and move forward
int	Ch;
[Ch:= Buffer(C2Inx);
if C2Inx<BufSize & ((Ch>=^A & Ch<=^Z) ! (Ch>=^a & Ch<=^z)) then
	[Ch:= Ch + (if Ch >=^a then -$20 else $20);
	Buffer(C2Inx):= Ch;
	FwdChar;
	Modified:= true;
	]
else	[FwdChar;
	SetStat("letter?", true);
	];
];	\FlipCase



proc	List;		\List edit Buffer on printer
int	C, Ch, Col;
[OpenO(PR);
FwdAll;
C:= 0;  Col:= 0;
while C < CInx do
	[Ch:= Buffer(C);  C:= C+1;

	if Ch >= $21 then [ChOut(PR, Ch);  Col:= Col+1]	\handle most characters
	else	[if Params(ShowFlag) then	\show normally hidden Ctrl chars
			[if Ch = Sp then ChOut(PR, SpInd)
			else	[ChOut(PR, ^^);	\convert Bel into "^G", etc.
				ChOut(PR, Ch+$40);
				];
			if Ch = CR then CrLf(PR);
			]
		else	[case Ch of
			  Tab:	[repeat	ChOut(PR, Sp);  Col:= Col+1;
				until	(Col & 7) = 0;
				];
			  FF:	[ChOut(PR, CR);  Col:= 0;  ChOut(PR, FF)];
			  CR:	[CrLf(PR);  Col:= 0]
			other	[ChOut(PR, Ch);  Col:= Col+1];
			];
		];
	];
];	\List

\-------------------------------------------------------------------------------

proc	XOn;		\Turn on scoop (aka: X register)
[ScoopOn:= true;
CutFlag:= false;
SetStat("scoop", false);		\(show it right away)
StatHeight:= 3;
];	\XOn



proc	XOff;		\Turn off scoop (aka: X register)
[ScoopOn:= false;
SetStat("ok", false);			\(show it right away)
];	\XOff



proc	CopyX;		\Copy scoop (X register) into the edit Buffer
\Note that the order of the text in the scoop is backwards.
int	X;
[ScoopOn:= false;			\make sure it's off
X:= XInx;
while X > 0 do
	[X:= X-1;
	Insert(Scoop(X));
	];
];	\CopyX



proc	BlockIns;	\Insert X-Register block mode
int	X, C, P;
[ScoopOn:= false;
X:= XInx;
if X > 0 then
   loop	[loop	[X:= X-1;
		if Scoop(X) = CR then
			[if C2Inx >= BufSize then Insert(CR);
			quit;
			];
		Insert(Scoop(X));
		if X < 1 then quit;
		];
	if X < 1 then quit;
	C:= DownLine;
	for P:= C to WorkCol do Insert(Sp);
	];
];	\BlockIns



proc	BlockFill;	\Fill a block of text along row/column lines with spaces
int	C1, C2, P1, P2;
[CutPtr2:= CInx;
if CutPtr1=CutPtr2 then return;

CutCol2:= RealCol;
if CutPtr1 > CutPtr2 then [P1:= CutPtr2;  P2:= CutPtr1]
else [P1:= CutPtr1;  P2:= CutPtr2];
if CutCol1 > CutCol2 then [C1:= CutCol2;  C2:= CutCol1]
else [C1:= CutCol1;  C2:= CutCol2];

MoveCur(P2);
repeat	GetColumn;
	if RealCol>=C1 & RealCol<=C2 & Buffer(CInx-1)#CR then
		[Buffer(CInx-1):= Sp;  Modified:= true];
	BackChar;
until	CInx = P1;
];	\BlockFill



proc	CutOn;		\Turn on text cutting
[CutFlag:= true;
ScoopOn:= false;
SetStat("cut", false);
StatHeight:= 3;
CutPtr1:= CInx;
CutCol1:= RealCol;
];	\CutOn



proc	CutText;	\Cutout a block of text along row/column lines
int	C1, C2, P1, P2;
[CutFlag:= false;
SetStat("ok", false);

CutPtr2:= CInx;
if CutPtr1 = CutPtr2 then return;

CutCol2:= RealCol;
if CutPtr1 > CutPtr2 then [P1:= CutPtr2;  P2:= CutPtr1]
else [P1:= CutPtr1;  P2:= CutPtr2];
if CutCol1 > CutCol2 then [C1:= CutCol2;  C2:= CutCol1]
else [C1:= CutCol1;  C2:= CutCol2];

MoveCur(P2);
repeat	GetColumn;
	ScoopOn:= RealCol>C1 & RealCol<=C2 ! Buffer(CInx-1)=CR;
	BackChar;
until	CInx = P1;
ScoopOn:= false;
];	\CutText

\-------------------------------------------------------------------------------

func	InsertHex(D);	\Insert literal hex value into edit Buffer
int	D;
int	H;


	func	Hex(Ch);	\Return the hex value
	int	Ch;		\hex digit in ASCII
	[case of
	  Ch>=^0 & Ch<=^9:	return Ch-$30;
	  Ch>=^A & Ch<=^F:	return Ch-$37;
	  Ch>=^a & Ch<=^f:	return Ch-$57
	other SetStat("hex?", true);
	];	\Hex


[	\InsertHex
H:= Hex(DBuf(DN, D));
H:= H*$10 + Hex(DBuf(DN, D+1));
if not ErrFlag then Insert(H);

return D+2;		\return index to next command
];	\InsertHex



func	InsertStr(D);	\Insert a string into edit Buffer
int	D;		\index into DBuf points to string
int	Ch;
[loop	[Ch:= DBuf(DN, D);
	if Ch = Esc then return D;	\return index to next command
	Insert(Ch);
	D:= D+1;
	];
];	\InsertStr

\-------------------------------------------------------------------------------

func	Search(D);	\Forward search edit Buffer for string
int	D;		\Index of first string character on command line (DBuf)
			\ or zero for string at edit Buffer cursor (C2Inx)
			\Returns index to next command (failed or not)
\Inputs: DBuf, DN, CInx, C2Inx
int	Ch1,		\first character in search string (for speed)
	S1F,		\Params(S1Flag), ignore case in searches
	Str0;		\index to start of search string (after any wild cards)
int	I, C, T;
char	SStr;		\search string



func	ToLower(Ch);	\Shift to lowercase if ignoring case in searches
int	Ch;		\ i.e: Params(S1Flag) is true
[if S1F then		\ignore case in searches
	if Ch>=^A & Ch<=^Z then		\shift to lowercase
		return Ch ! $20;
return Ch;
];	\ToLower



func	NotAlphaNum(Ch);	\Return "true" for punctuation i.e.
int	Ch;			\ non-alpha-numeric
[return Ch<^0 ! (Ch>^9 & Ch<^A) ! (Ch>^Z & Ch<^a) ! Ch>^z;
];	\NotAlphaNum



proc	FwdSearch;	\Forward search edit Buffer for string.
[C:= C2Inx + Str0;			\start searching from the cursor
T:= 0;
I:= Str0;				\point to first non-wild char in string
loop	[case SStr(I) of
	  WildCh, ToLower(Buffer(C+T)):
		[			\it matches
		I:= I+1;		\try next string character
		T:= T+1;
		if SStr(I) = Esc then	\the entire string matches
			[C:= C+T;		\point to end of string
			if C > BufSize then	\(ignore garbage beyond buffer)
				[SetStat("none?", true);  quit];

			\Move cursor forward to end of matching string:
			CopyBuffer(C2Inx, CInx, C-C2Inx);
			CInx:= CInx + C - C2Inx;
			C2Inx:= C;
			SavePtrs;
			quit;
			];
		];
	PunctCh:
	    if NotAlphaNum(Buffer(C+T)) then
		[			\it matches
		I:= I+1;		\try next string character
		T:= T+1;
		if SStr(I) = Esc then	\the entire string matches
			[C:= C+T;		\point to end of string
			if C > BufSize then	\(ignore garbage beyond buffer)
				[SetStat("none?", true);  quit];

			\Move cursor forward to end of matching string:
			CopyBuffer(C2Inx, CInx, C-C2Inx);
			CInx:= CInx + C - C2Inx;
			C2Inx:= C;
			BackChar;	\Stop on end of word, not punctuation
			SavePtrs;
			quit;
			];
		]
	    else  [			\it doesn't match
		  C:= C+1;		\move forward in edit Buffer
		  if C >= BufSize then	\all done? then not found
			[SetStat("none?", true);  quit];

		  \Fast scan for first character (Ch1):
		  \Wild characters are ignored
		  loop	[if Ch1 = PunctCh then
			    for C:= C to BufSize do
				[if NotAlphaNum(Buffer(C)) then quit]
			else
			  if S1F then	\if ignore case in searches
			    for C:= C to BufSize do
				[if (Buffer(C) ! $20) = Ch1 then quit]
				\(good candidate--it's rechecked above)
			  else
			    for C:= C to BufSize do
				if Buffer(C) = Ch1 then quit;
			  quit;		\quit if end of buffer
			];
		  I:= Str0;		\point to first char in string
		  T:= 0;
		  ]
	other	[			\it doesn't match
		C:= C+1;		\move forward in edit Buffer
		if C >= BufSize then	\all done? then not found
			[SetStat("none?", true);  quit];

		\Fast scan for first character (Ch1):
		\Wild characters are ignored
		loop	[if Ch1 = PunctCh then
			    for C:= C to BufSize do
				[if NotAlphaNum(Buffer(C)) then quit]
			else
			if S1F then	\if ignore case in searches
			    for C:= C to BufSize do
				[if (Buffer(C) ! $20) = Ch1 then quit]
				\(good candidate--it's rechecked above)
			else
			    for C:= C to BufSize do
				if Buffer(C) = Ch1 then quit;
			quit;		\quit if end of buffer
			];

		I:= Str0;		\point to first char in string
		T:= 0;
		];
	];	\loop
];	\FwdSearch



proc	BackSearch;	\Backward search edit Buffer for string.
[C:= CInx-I-1;			\searching from the cursor - string length -1
I:= Str0;			\point to first char in string
T:= 0;
loop	[case SStr(I) of
	  WildCh, ToLower(Buffer(C+T)):
		[			\it matches
		I:= I+1;		\try next string character
		T:= T+1;
		if SStr(I) = Esc then	\the entire string matches
			[if C-Str0 < 0 then	\(ignore garbage in front)
				[SetStat("none?", true);  quit];
			C:= C+T;		\point to end of string

			\Move cursor backward to end of matching string:
			CopyBuffer(C, C2Inx-(CInx-C), CInx-C);

			C2Inx:= C2Inx - (CInx-C);
			CInx:= CInx - (CInx-C);
			SavePtrs;
			quit;
			];
		];
	PunctCh:
	    if NotAlphaNum(Buffer(C+T)) then
		[			\it matches
		I:= I+1;		\try next string character
		T:= T+1;
		if SStr(I) = Esc then	\the entire string matches
			[if C-Str0 < 0 then	\(ignore garbage in front)
				[SetStat("none?", true);  quit];
			C:= C+T;		\point to end of string

			\Move cursor backward to end of matching string
			CopyBuffer(C, C2Inx-(CInx-C), CInx-C);

			C2Inx:= C2Inx - (CInx-C);
			CInx:= CInx - (CInx-C);
			BackChar;	\stop on end of word, not punctuation
			SavePtrs;
			quit;
			];
		]
	    else  [			\it doesn't match
		  C:= C-1;		\move backward in edit Buffer
		  if C < 0 then		\all done? then not found
			[SetStat("none?", true);  quit];

		  \Fast scan for first character (Ch1):
		  \Wild characters are ignored
		  loop	[if Ch1 = PunctCh then
			    for C:= C downto 0 do
				[if NotAlphaNum(Buffer(C)) then quit]
			else if S1F then \if ignore case in searches
			    for C:= C downto 0 do
				[if (Buffer(C) ! $20) = Ch1 then quit]
				\(good candidate--it's rechecked above)
			else for C:= C downto  0 do
				if Buffer(C) = Ch1 then quit;
			quit;		\quit if start of buffer
			];
		  I:= Str0;		\point to first char in string
		  T:= 0;
		  ]
	other	[			\it doesn't match
		C:= C-1;		\move backward in edit Buffer
		if C < 0 then		\all done? then not found
			[SetStat("none?", true);  quit];

		\Fast scan for first character (Ch1):
		\Wild characters are ignored
		loop	[if Ch1 = PunctCh then
			    for C:= C downto 0 do
				[if NotAlphaNum(Buffer(C)) then quit]
			else if S1F then \if ignore case in searches
			    for C:= C downto 0 do
				[if (Buffer(C) ! $20) = Ch1 then quit]
				\(good candidate--it's rechecked above)
			else for C:= C downto 0 do
				if Buffer(C) = Ch1 then quit;
			quit;		\quit if start of buffer
			];
		I:= Str0;		\point to first char in string
		T:= 0;
		];
	];	\loop
];	\BackSearch



[	\Search
SStr:= Reserve(DBufSize);
if D#0 then			\get string from deferred command buffer
	[\Check for null string (no can handle)
	if DBuf(DN,D) = Esc then [SetStat("null?", true);  return D];

	S1F:= Params(S1Flag);			\ignore case in searches?
	I:= 0;  Str0:= -1;
	loop	[SStr(I):= ToLower(DBuf(DN,D));	\work with a copy of the string
		if SStr(I) = Esc then quit;
		if SStr(I) # WildCh then if Str0<0 then Str0:= I; \point to 1st
								\ non-wild card
		I:= I+1;
		D:= D+1;
		];
	\Handle string with all wild cards
	if Str0 = -1 then Str0:= 0;
	]
else	[			\get string from edit Buffer
	\Move the cursor to the start of the word
	loop	[				\back over the word
		if NotAlphaNum(Buffer(CInx-1)) ! ErrFlag then quit;
		BackChar;
		if CInx = 0 then quit;
		];
	C:= C2Inx;				\point to start of string

	S1F:= Params(S1Flag);			\ignore case in searches?
	I:= 1;
	loop	[SStr(I):= ToLower(Buffer(C));	\work with a copy of the string
		if NotAlphaNum(Buffer(C)) ! ErrFlag ! C >=BufSize then quit;
		I:= I+1;
		C:= C+1;
		];
	SStr(0):= PunctCh;  SStr(I):= PunctCh;	\surround with non-alpha-numeric
	SStr(I+1):= Esc;			\terminate the string
	Str0:= 0;				\no wild cards in edit Buffer
	];

Ch1:= SStr(Str0);		\get first non-wild character in search string
if S1F then Ch1:= Ch1 ! $20;	\shift to lowercase
if Params(S2Flag) then BackSearch else FwdSearch;
\if not ScrollMode then HeightCur:= FixedCur;
return D;
];	\Search

\-------------------------------------------------------------------------------

func	DirSearch(C, D); \Do search, change direction based on C
int	C, D;
[case C of
  ^+:	Params(S2Flag):= false;
  ^-:	Params(S2Flag):= true
other	[];
return Search(D);
];	\DirSearch



func	ReplaceStr(D);	\Replace one string with another
int	D;
int	DX, Ch, I, SP;
def	StkSize = 128;	\maximum wild cards (^W's and ~W's) handled
char	Stk(StkSize);
[DX:= Search(D);			\find string
if ErrFlag then
	repeat DX:= DX+1 until DBuf(DN,DX) = Esc
else	[SP:= 0;
	if DBuf(DN,DX-1) = \PunctCh\-AltW then FwdChar;
	for I:= DX-1 downto D do	\delete it
		[DelChar;
		Ch:= DBuf(DN,I);
		if Ch=WildCh ! Ch=\PunctCh\-AltW then
		    if SP < StkSize then
			[Stk(SP):= Buffer(CInx);  SP:= SP+1];
		];
	DX:= DX+1;			\replace it
	loop	[Ch:= DBuf(DN,DX);
		if Ch = Esc then quit;
		if Ch=WildCh ! Ch=\PunctCh\-AltW then
		    if SP > 0 then
			[SP:= SP-1;  Ch:= Stk(SP)];
		Insert(Ch);
		DX:= DX+1;
		];
	];
return DX;
];	\ReplaceStr

\-------------------------------------------------------------------------------

proc	Indent;		\Indent current line same as the preceding one
int	C;		\temporary index into text buffer
[\Move to the beginning of the previous line:
if Buffer(CInx-1) = CR then BackLine
else	[BackLine;  BackLine];

C:= C2Inx;		\point to start of line before

if not ErrFlag then FwdLine
else SetStat("ok", false);

\Delete all initial tabs and spaces on the current line:
loop	[case Buffer(C2Inx) of
	  Sp, Tab: if C2Inx < BufSize then [FwdChar;  DelChar] else quit
	other quit;
	];

\Copy exact margin from line above
loop	case Buffer(C) of
	  Tab, Sp: [Insert(Buffer(C));  C:= C+1]
	other quit;
\(Leave cursor at the first displayable character)
];	\Indent



proc	Center;		\Center the current line between margins
int	Col,		\The column number to indent to
	C,		\temporary index into text buffer
	I;		\scratch
[\Move to the beginning of the line:
if Buffer(CInx-1)#CR & CInx>0 then BackLine;

\Delete all initial tabs and spaces on the line:
loop	[case Buffer(C2Inx) of
	  Sp, Tab: if C2Inx < BufSize then [FwdChar;  DelChar] else quit
	other quit;
	];

\Count the number of printable characters remaining on the line:
C:= C2Inx;
Col:= 0;
loop	[if Buffer(C)=CR ! C>=BufSize then quit;
	if Buffer(C)>=Sp & Buffer(C)<=$FE then Col:= Col+1;
	C:= C+1;
	];

\Calculate amount to indent line:
Col:= (Params(RightMargin) - Params(LeftMargin) - Col + 1) / 2;
Col:= Col + Params(LeftMargin);

\Insert spaces into edit Buffer:
for I:= 1 to Col do Insert(Sp);
\(Leave cursor at the first displayable character)
];	\Center

\-------------------------------------------------------------------------------

proc	NextMark;	\Advance to next marker
MarkPtr:= if MarkPtr < 1 then 1 else 0;



proc	DropMark;	\Save a marker of the current location
[Marker(MarkPtr):= CInx;
NextMark;
];	\DropMark



proc	LastMark;	\Move cursor to last location
[NextMark;
if not MoveCur(Marker(MarkPtr)) then SetStat("range?", true);
];	\LastMark

\-------------------------------------------------------------------------------

proc	DoMargin;	\Do actual re-margining
int	WT,		\flag: wordwrap mode
	OT,		\hold overwrite flag
	Col,		\column
	I;		\scratch
[WT:= Params(WrapFlag);		\wordwrap mode must be off
Params(WrapFlag):= false;
OT:= Params(OverFlag);		\overwrite must be off
Params(OverFlag):= false;

Buffer(BufSize):= 0;		\kluge to fix numerous bugs in following code

if CInx > 0 then BackChar;	\(margin back)

\Move to the end of the paragraph (2 CR's or end of buffer):
loop	[if C2Inx >= BufSize then quit;
	if Buffer(C2Inx) = CR then
		[FwdChar;
		\Discard (hidden) spaces and tabs between paragraphs:
		while (Buffer(C2Inx)=Sp ! Buffer(C2Inx)=Tab) &
			C2Inx<BufSize do FwdChar;
		if Buffer(C2Inx)=CR ! C2Inx>=BufSize then
			[while Buffer(CInx-1) # CR do DelChar;	\(tricky)
			BackChar;  BackChar;
			quit;
			];
		];
	FwdChar;
	];
	\C2Inx points to last character in the paragraph
SetStat("ok", false);		\(in case there are 2 CR's at the beginning)

\Move to the start of the paragraph replacing CR's and tabs with spaces:
loop	[if Buffer(C2Inx) = Tab then Buffer(C2Inx):= Sp;
	if Buffer(C2Inx) = CR then
		[if CInx <= 0 then [FwdChar;  quit];
		BackChar;
		if Buffer(C2Inx) = CR then [FwdChar;  FwdChar;  quit];
		FwdChar;
		Buffer(C2Inx):= Sp;
		];
	if CInx <= 0 then quit;
	BackChar;
	];
SetStat("ok", false);

\Move to the end of the paragraph deleting multiple spaces:
while Buffer(C2Inx) = Sp do		\delete any spaces at the beginning
	[FwdChar;  DelChar];
loop	[if Buffer(C2Inx) = Sp then
		[FwdChar;
		while Buffer(C2Inx) = Sp do
			[FwdChar;  DelChar];
		];
	if Buffer(C2Inx)=CR ! C2Inx>=BufSize then quit;
	FwdChar;
	];
if Buffer(CInx-1) = Sp then DelChar;	\remove space from end of paragraph
SetStat("ok", false);

\Move to the start of the paragraph (now only one line long):
BackLX;
SetStat("ok", false);

Col:= Params(LeftMargin) + Params(ParaIndent);	\the first line can be indented
loop	[\insert tabs and spaces into edit Buffer to get to the left margin:
	for I:= 1 to Col do Insert(Sp);
	Insert(CR);			\insert a fake CR to mark our place
	\(This is required if we have a word longer than a line and we are
	\ indenting with spaces.)

	\Move to the right margin +1:
	loop	[if Col > Params(RightMargin) then quit;
		FwdChar;		\(sets ErrFlag at end of buffer)
		if Buffer(C2Inx) = CR ! ErrFlag then	\all done
			[SetStat("ok", false);		\in case of ErrFlag
			BackLX;				\back up to fake CR
			DelChar;			\eat it
			FwdLX;				\move to new line
			FwdChar;
			Params(WrapFlag):= WT;		\restore wordwrap mode
			Params(OverFlag):= OT;		\restore overwrite
			return;
			];

		if Buffer(C2Inx)>=$20 & Buffer(C2Inx)<=$FE then
			Col:= Col+1;	\don't count control chars
		];

	\Move back to previous space and replace it with a CR:
	loop	[if Buffer(C2Inx) = Sp then
			[Buffer(C2Inx):= CR;  quit];

		\Deal with case where a word is longer than a line:
		if Buffer(C2Inx) = CR then	\(our fake CR)
			repeat	[FwdChar;
				if Buffer(C2Inx) = Sp then
					[Buffer(C2Inx):= CR;  quit];
				if Buffer(C2Inx) = CR ! ErrFlag then quit;
				];
			until	false;
		BackChar;
		if ErrFlag then quit;
		];

	if ErrFlag then quit;
	BackLX;			\back up to fake CR
	DelChar;		\eat it
	FwdLX;			\move to new line
	Col:= Params(LeftMargin);
	];
Params(WrapFlag):= WT;		\restore wordwrap mode
Params(OverFlag):= OT;		\restore overwrite
];	\DoMargin



func	Margin(D);	\Re-margin paragraph (or get bell column)
int	D;		\index to arguments (if any)
int	LM, RM,		\left & right margin columns
	BellFlag;	\this is a bell set command (not a margin)


	func	NumIn;
	\Get positive integer from command line and return its value
	\Also moves D to the char terminating the number
	int	Num, Ch;
	[Num:= 0;
	Ch:= DBuf(DN, D);
	while Ch = Sp do		\eat leading spaces
		[D:= D+1;
		Ch:= DBuf(DN, D);
		];
	while Ch>=^0 & Ch<=^9 do
		[Num:= Num*10 + (Ch -^0);
		if Num < 0 then [SetStat("number?", true);  return 0];
		D:= D+1;
		Ch:= DBuf(DN, D);
		];
	return Num;
	];	\NumIn


	proc	GetArgs;	\Get arguments to margin command
	[				\we have arguments--go get 'em
	if DBuf(DN,D) =^( then D:= D+1;	\parentheses are optional
	LM:= NumIn;			\get the new left margin column
	BellFlag:= false;		\assume it is not a bell setting
	if DBuf(DN,D)=^, ! DBuf(DN,D)=Sp then [D:= D+1;  RM:= NumIn]
	else BellFlag:= true;
	if DBuf(DN,D)=^, ! DBuf(DN,D)=Sp then
		[D:= D+1;  Params(ParaIndent):= NumIn];
	if DBuf(DN,D) = ^) then D:= D+1;\parentheses are optional
	if BellFlag then		\only one argument means set bell
		[Params(BellCol):= LM;  return D+1];	\ and we're done
	Params(LeftMargin):= LM;	\set new left and right margin columns
	Params(RightMargin):= RM;
	];	\GetArgs


[	\Margin
if DBuf(DN,D)>=^0 & DBuf(DN,D)<=^9 ! DBuf(DN,D)=^( then [GetArgs;  return D];
DoMargin;
return D;
];	\Margin



proc	ConSetup;	\Setup to interpret the text in scoop as a number
\Note that the order of the text in scoop is backwards
int	P, C;
def	XMax = 80;
[OpenO(8);
OpenI(8);
C:= if XInx < XMax then XInx else XMax;
for P:= C-1 downto 0 do ChOut(8, Scoop(P));
ChOut(8, CR);
];	\ConSetup



proc	ConX2Dec;	\Interpret text in scoop as a decimal number
[ConSetup;
Global:= IntIn(8);
];	\ConX2Dec



proc	ConX2Hex;	\Interpret text in scoop as a hex number
[ConSetup;
Global:= HexIn(8);
];	\ConX2Dec



proc	ShowDec;	\Display Global value on status line in decimal
int	P;
[OpenO(8);  OpenI(8);
IntOut(8, Global);
ChOut(8, EOF);		\for safety
P:= -1;
repeat P:= P+1;  HexStr(P):= ChIn(8) until HexStr(P) = EOF;
if P > 0 then HexStr(P-1):= HexStr(P-1) ! $80 else HexStr(0):= $A0;
SetStat(HexStr, false);
];	\ShowDec



proc	ShowHex;	\Display Global as hex on status line
int	P, S;
[OpenO(8);  OpenI(8);
HexOut(8, Global);
S:= 7;
for P:= 0 to S do HexStr(P):= ChIn(8);
HexStr(S):= HexStr(S) ! $80;
SetStat(HexStr, false);
];	\ShowHex



proc	DecInsert;	\Insert Global into edit Buffer as a decimal string
int	C;
[OpenO(8);
OpenI(8);
IntOut(8, Global);
ChOut(8, EOF);		\for safety
loop	[C:= ChIn(8);
	if C = EOF then quit;
	Insert(C);
	];
];	\DecInsert



proc	HexInsert;	\Insert Global into edit Buffer as a hex string
int	C;
[OpenO(8);
OpenI(8);
HexOut(8, Global);
ChOut(8, EOF);		\for safety
loop	[C:= ChIn(8);
	if C = EOF then quit;
	Insert(C);
	];
];	\HexInsert



func	GlobalCmd(D);	\Commands to display or insert Global
int	D;
[case DBuf(DN, D) of
  ^X,^x: ConX2Hex;
  ^T,^t: ConX2Dec;
  ^N,^n: ShowDec;
  ^S,^s: ShowHex;
  ^A,^a: Insert(Global);
  ^D,^d: DecInsert;
  ^H,^h: HexInsert
other	 [];
return D+1;
];	\GlobalCmd

\================================ I/O COMMANDS =================================

func	LowOpen(Str, Mode);	\Open file, return handle. Handle=0 if error.
char Str;  int Mode;
int	Hand;
[Trap(false);
Hand:= FOpen(Str, Mode);
Trap(true);
return if GetErr=0 then Hand else 0;
];	\LowOpen



proc	AbandonIO;	\Abandon input and output files
[if InHand then FClose(InHand);
if OutHand then FClose(OutHand);
asm	{ldr	r0, OutTemp		@ delete outfile.bak
	bl	remove			@ (.bak file might not exist)
	}
];	\AbandonIO



proc	CopyName(From, To, Ext);
\Copy file name from From to To and replace any existing extension with Ext
char	From, To,	\Strings are terminated by 0 & must be <= NameSize long
	Ext;		\Extension started with "." and terminated with an
			\ unused char with its MSB set
int	I, J, Ch;
[I:= 0;
loop	[Ch:= From(I);
	if Ch=^. ! Ch=0 then quit;
	To(I):= Ch;
	if I < NameSize-1 then I:= I+1;
	];
for J:= 0 to 3 do
	[To(I):= Ext(J);
	if I < NameSize-1 then I:= I+1;
	];
To(I):= 0;
];	\CopyName



func	OpenAll;	\Open input and files based on OutFile and InFile
int	P;
[if OutFile(0) = 0 then			\if no outfile then use the infile
	for P:= 0 to NameSize-1 do
		OutFile(P):= InFile(P);
CopyName(OutFile, OutTemp, ".tmp ");	\set up other output file names
CopyName(OutFile, OutBack, ".bak ");

Trap(false);
InHand:= LowOpen(InFile, 0\read\);	\get input and output handles

\If no input file then set flag so we don't attempt to read from it
OutHand:= LowOpen(OutTemp, 1\write\);
if OutHand=0 then return false;
Trap(true);

if InHand then FSet(InHand, ^I);
FSet(OutHand, ^O);
return true;
];	\OpenAll



proc	DOSOpen;	\Routine to open (MS-DOS) files
int	Inx, Ch;	\index into Tail
char	Tail;


	proc	GetName(Name);	\Inputs: Inx, Tail, NameSize
	char	Name;
	int	I, Ch;
	[while Tail(Inx) = Sp do Inx:= Inx+1;
	I:= 0;
	repeat	Ch:= Tail(Inx);
		Inx:= Inx+1;
		if Ch=Sp ! Ch=CR ! Ch=EOF then Ch:= 0;	\terminate Name with 0
		Name(I):= Ch;
		if I < NameSize-1 then I:= I+1;
	until	Ch = 0;
	];	\GetName


[	\DOSOpen
Tail:= Reserve($80);			\get command tail
Inx:= 0;
repeat	Ch:= ChIn(8);
	Tail(Inx):= Ch;
	Inx:= Inx+1;
until	Ch=0 ! Ch=EOF ! Inx>=$80;

Inx:= 0;
GetName(InFile);			\fatal error if no file specified
if InFile(0) = 0 then
	[SetVid(3);
	Text(0, "No output file specified");
	CrLf(0);
	exit 1;
	];

Inx:= Inx-1;
GetName(OutFile);

if not OpenAll then
	[SetVid(3);  Text(0, "Bad output file");  CrLf(0);  exit 1];
];	\DOSOpen



proc	DOSClose;	\Close "MS-DOS disk" files
[Trap(false);
if OutHand then FClose(OutHand);
OutHand:=0;
if InHand then FClose(InHand);
InHand:=0;

asm	{ldr	r0, OutBack		@ delete outfile.bak
	bl	remove			@ (.bak file might not exist)

	ldr	r0, OutFile		@ rename outfile.ext to outfile.bak
	ldr	r1, OutBack
	bl	rename

	ldr	r0, OutTemp		@ rename outfile.tmp to outfile.ext
	ldr	r1, OutFile
	bl	rename
	}

Trap(true);
if GetErr then SetStat("file?", true);
];	\DOSClose



proc	CloseFile;
[\ChOut(3, EOF);
Close(3);
DOSClose;
];	\CloseFile

\-------------------------------------------------------------------------------

proc	OpenFiles;	\Open the files
[DOSOpen;		\open (MS-DOS) files
Trap(false);
if InHand then OpenI(3);
OpenO(3);
BaseLine:= 0;
Trap(true);
if GetErr then SetStat("file?", true);
EOFFlag:= InHand = 0;
];	\OpenFiles



proc	Rewind;		\Rewind files
[Trap(false);
if InHand then OpenI(3);
OpenO(3);
BaseLine:= 0;
Trap(true);
if GetErr then SetStat("file?", true)
else EOFFlag:= InHand = 0;
];



func	OCmd(D);	\Handle "O" command
int	D;
int	P;
[if DBuf(DN,D)=^" & D<DInx(DN) then
	[P:= 0;
	loop	[D:= D+1;
		if D > DInx(DN) then [SetStat("no ^"", true);  return D];
		InFile(P):= DBuf(DN, D);
		if InFile(P) = ^" then [InFile(P):= 0;  quit];
		P:= P+1;
		if P>=NameSize ! D>DInx(DN) then
			[SetStat("no ^"", true);  return D];
		];
	D:= D+1;
	Rewind;
	AbandonIO;
	OutFile(0):= 0;
	if not  OpenAll then [SetStat("outfile?", true);  return D];
	Trap(false);
	if InHand then OpenI(3);
	OpenO(3);
	BaseLine:= 0;
	Trap(true);
	if GetErr then SetStat("file?", true);
	EOFFlag:= InHand=0;
	]
else	Rewind;
return D;
];	\OCmd



func	ReadChar;	\Read a character from input file
int	Ch;
[if HaveCR then
	repeat Ch:= ChIn(3) until Ch#LF	\convert CR+LF to CR
else	[Ch:= ChIn(3);			\convert LF to CR
	if Ch = CR then HaveCR:= true
	else if Ch = LF then Ch:= CR;
	];
return Ch;
];	\ReadChar



proc	WriteChar(Ch);	\Write a character to output file
int	Ch;
[if HaveCR then
	[ChOut(3, Ch);			\convert CR to CR+LF
	if Ch = CR then ChOut(3, LF);
	]
else	[if Ch = CR then Ch:= LF;	\convert CR to LF
	ChOut(3, Ch);
	];
];	\WriteChar



proc	Append;		\Append input file to edit Buffer
int	Ch, Limit, T;
def	LLimit = 256;	\limit looking for last CR

	proc	ReadFile;	\Read file, leaving Line Feeds
	int	P;
	[HaveCR:= false;		\last file appended determines format
	for CInx:= CInx to Limit do	\read first part of file fast
		[Ch:= ReadChar;
		case of
		  GetErr: [SetStat("infile?", false);  Trap(true);  return];
		  Ch=EOF: [EOFFlag:= true;  SetStat("eof", false);
			   Trap(true);  return]
		other	  [];
		Buffer(CInx):= Ch;
		];
	for P:= 1 to LLimit do		\read last part, looking for CR to stop
		[if CInx >= C2Inx then [SetStat("full?", true);  return];
		Ch:= ReadChar;
		case of
		  Ch=EOF: [EOFFlag:= true;  SetStat("eof", false);  return];
		  GetErr: [SetStat("infile?", true);  return]
		other	  [];
		Buffer(CInx):= Ch;  CInx:= CInx+1;
		if Ch = CR then return;	\stop at the end of a line
		];
	];	\ReadFile

[	\Append
T:= if Pad > (XInx+1024) then Pad else XInx+1024;
Limit:= C2Inx - T;
Trap(false);
if EOFFlag then SetStat("eof?", true) else ReadFile;
SavePtrs;
Trap(true);
];	\Append



proc	Write2Cursor;	\Write chars from start to cursor
int	I, Ch;
[BaseLine:= BaseLine + LineCount(Buffer, CInx);

Trap(false);
for I:= 0 to CInx-1 do
	[Ch:= Buffer(I);
	WriteChar(Ch);
	if Ch = CR then
	    if GetErr then
		[SetStat("outfile?", true);
		Trap(true);
		return false;
		];
	];
Trap(true);
return true;
];	\Write2Cursor



proc	WriteBuf;	\Write the edit Buffer to the output file
[if OutHand then
	[FwdAll;
	Write2Cursor;
	]
else	SetStat("closed?", true);
];	\WriteBuf



proc	NextBuf;	\Get next buffer-full (Esc G)
[WriteBuf;
if not ErrFlag then
	[DelAll;
	Append;
	];
];	\NextBuf



proc	Shuffle(N);	\Read and write N lines into and out of editor
int	N, Ch, C;


	proc	MoveLines;	\Write and delete first line, read in last line
	int	P;
	[ScoopOn:= false;
	CutFlag:= false;
	BackAll;
	FwdLine;
	Write2Cursor;
	CInx:= 0;
	FwdAll;
	P:= 0;
	loop	[		\read last part, looking for CR to stop on
		if CInx >= C2Inx then [SetStat("full?", true);  quit];
		Ch:= ReadChar;
		case of
		  Ch=EOF: [EOFFlag:= true;  SetStat("eof", false);  quit];
		  GetErr: [SetStat("infile?", true);  quit]
		other	  [];
		Buffer(CInx):= Ch;  CInx:= CInx+1;  P:= P+1;
		\stop at the end of a line or at limit
		if Ch = CR ! P>256 then quit;
		];
	];	\MoveLines


[	\Shuffle
Trap(false);
C:= 1;
loop	[if EOFFlag then [SetStat("eof?", true);  quit];
	MoveLines;
	C:= C+1;
	if C > N then quit;
	];
SavePtrs;
Trap(true);
];	\Shuffle



func	BufSearch(D);	\Forward search N buffers for string
int	D;
int	DX, S2F;
[S2F:= Params(S2Flag);			\save current search direction
Params(S2Flag):= false;			\set to forward search
loop	[DX:= Search(D);
	if ErrFlag then			\we didn't find it in this buffer
		[if not EOFFlag then
			[SetStat("ok", false);	\clear ErrFlag from search
			NextBuf;
			SetStat("ok", false);	\clear possible eof flag
			BackAll]
		else quit;		\(leaves last buffer in memory)
		]
	else	quit;			\found it, stop looking
	];
Params(S2Flag):= S2F;			\restore search direction
return DX;				\return index to next command
];	\BufSearch

\-------------------------------------------------------------------------------

proc	SaveConfig;	\Save configuration data
\Inputs:  DInx, DBuf, DN
int	Hand, I, J, Len;
begin
Trap(false);				\don't abort if an error is detected
Hand:= FOpen(".xed-config.ini", 1);	\get output file handle
Trap(true);
if GetErr \#0\ then []			\if error detected then ignore request
else	begin
	FSet(Hand, ^o);
	OpenO(3);
	IntOut(3, DN);  CrLf(3);	\currently selected command line
	for J:= 0 to DNSize-1 do	\for all command lines...
		[Len:= DInx(J);
		IntOut(3, Len);  CrLf(3);
		for I:= 0 to Len-1 do ChOut(3, DBuf(J,I));
		CrLf(3);
		];
	Close(3);
	FClose(Hand);
	end;
end;	\SaveConfig



proc	LoadConfig;	\Load configuration data file
\Outputs:  DInx, DBuf, DN
int	Hand, I, J, Len;
begin
Trap(false);				\don't trap "file not found" error
Hand:= FOpen(".xed-config.ini", 0);	\open file for input
Trap(true);
if GetErr \#0\ then []			\if error detected then ignore request
else	begin
	FSet(Hand, ^i);			\assign device 3 to handle
	OpenI(3);
	DN:= IntIn(3);			\restore last selected command line
	if DN<10 ! DN>DNSize then DN:= 10;
	for J:= 0 to DNSize-1 do
		[Len:= IntIn(3);
		if Len > DBufSize then Len:= DBufSize;
		DInx(J):= Len;
		I:= ChIn(3);		\eat LF
		for I:= 0 to Len-1 do DBuf(J,I):= ChIn(3);
		];
	FClose(Hand);
	end;
end;	\LoadConfig

\-------------------------------------------------------------------------------

proc	Exit;		\Exit from XED
[SetVid(3);		\restore original screen depth
Text(0, "                              -- XED 3.4 --

             Based on the original version by P.J.R. Boyle.
                        Copyright (c) 1984-2018
             Richard Ottosen, Loren Blaney, and Larry Fish.

");
exit 0;
];	\Exit



proc	Quit;		\Output the buffer and any remaining input file
int	Ch;
[if OutHand then				\save file
	[WriteBuf;
	if ErrFlag then return;			\no exit if any error
	if not EOFFlag then			\if EOF is not encountered then
		[Trap(false);			\ shuffle in file to out file
		Ch:= ReadChar;
		WriteChar(Ch);
		loop	[Ch:= ReadChar;
			if Ch = EOF then quit;
			if GetErr then
				[SetStat("file?", true);  Trap(true);  return];
			WriteChar(Ch);
			];
		EOFFlag:= true;			\(in case we don't exit)
		];
	\ChOut(3, EOF);
	Close(3);
	Trap(true);
	DOSClose;
	];
SaveConfig;
Exit;
];	\Quit



proc	Terminate;	\Terminate editor and "Trash" (Esc T) output file
int	X, Y;		\But first verify before discarding a modified file
[if Modified then
  loop	[Attrib(StatAtt);
	X:= (Width-16)/2;  Y:= (Height-5)/2;
	SetWind(X, Y, X+16-1, Y+5-1, 2\clip\, 0);
	RawText(6, "Ŀ
 Save changes 
              
   (Y/N)?     
 ");
	\position red cursor
	CurH:= (Width-16)/2 + 11;
	CurL:= (Height-5)/2 + 3;
	LeftCol:= 0;
	case KeyIn($0C) of
	  ^Y, ^y:	Quit;		\save changes
	  ^N, ^n:	quit;		\continue with "Trash" command
	  Esc:		[SetWind(0, 0, 1000, 1000, 0, 0);
			Display;	\restore edit screen
			return]
	other		[];		\ignore illegal command
	];
Trap(false);		\ignore any I/O errors
AbandonIO;
Trap(true);
X:= GetErr;
SaveConfig;
Exit;
];	\Terminate

\============================ EXPRESSION EVALUATOR =============================

func	Eval(EPtr);	\Evaluate expression pointed to by D
int	EPtr;		\Index into string containing expression on command line



proc	GetCh;
[if EPtr >= DInx(DN) then
	[SetStat("express?", true);
	Char:= ^);
	]
else	[Char:= DBuf(DN, EPtr);
	EPtr:= EPtr+1;
	];
];	\GetCh



proc	SkipTab;	\Skip tabs and spaces in DBuf
[while Char=Tab ! Char=Sp do GetCh;
];	\SkipTab



proc	GetChX;		\Get next character, ignoring any space characters
[repeat GetCh until Char # Sp;
];	\GetChX



func	UnderCur;	\Get character under cursor
int	Val;
[if C2Inx < BufSize then Val:= Buffer(C2Inx)
else [SetStat("end?", true);  Val:= 0];
GetCh;
return Val;
];	\UnderCur



func	GetASCII;
int	Val;
[GetCh;
Val:= Char;
GetCh;
return Val;
];	\GetASCII



func	GetDec;		\Read in a decimal string and return its value
int	Val;		\Terminate on first non-number.
[Val:= Char-^0;
GetCh;
if Char>=^a & Char<=^z then Char:= Char-$20;	\to uppercase
loop	[if Char<^0 ! Char>^9 then quit;
	Val:= Val*10 + Char-^0;
	GetCh;
	];
return Val;
];	\GetDec



func	GetHex;		\Return the value of a hex string, e.g: $000012AB
int	Val, I;
[GetCh;						\skip the "$"
if Char>=^a & Char<=^z then Char:= Char-$20;	\to uppercase
case of
  Char>=^0 & Char<=^9: Val:= Char-^0;
  Char>=^A & Char<=^F: Val:= Char-$37
other [SetStat("hex?", true);  return 0];
loop	[GetCh;
	if Char>=^a & Char<=^z then Char:= Char-$20;
	case of
	  Char>=^0 & Char<=^9: I:= Char-^0;
	  Char>=^A & Char<=^F: I:= Char-$37
	other quit;
	Val:= Val<<4 + I;
	];
return Val;
];	\GetHex

\-------------------------------------------------------------------------------

func	Expression;	\Returns the value of an expression
int	Val;



func	Factor;		\Returns the value of a factor
int	Val, Neg;
[while Char = ^+ do GetChX;		\ignore unary "+"
Neg:= false;
while Char =^- do			\unary "-"
	[GetChX;  Neg:= not Neg];

if Char>=^0 & Char<=^9 then		\unsigned integer
	Val:= GetDec
else case Char of
  ^(:	[GetChX;			\sub-expression
	Val:= Expression;
	if Char#^) then SetStat("express?", true);
	GetChX;
	];
  ^K,^k: [Display;  OpenI(0);  Val:= ChIn(1);  GetCh];
  ^Z,^z: [Val:= C2Inx - CInx;  GetCh];	\get buffer free space
  ^C,^c: [GetColumn;  Val:= RealCol;  GetCh]; \get value of global variable
  ^%:	[Val:= Global;  GetCh];		\get value of global variable
  ^$:	Val:= GetHex;			\unsigned hex integer
  ^^:	Val:= GetASCII;			\string constant
  ^@:	Val:= UnderCur;			\character under cursor
  ^.:	[Val:= CInx;  GetChX]		\cursor position
other	SetStat("syntax?", true);

SkipTab;	\identifiers, numbers, etc. may terminate with a space or tab
return if Neg then -Val else Val;
];	\Factor



func	Term;		\Returns Factor * Factor, etc.
int	Val;
[Val:= Factor;
loop	case Char of
	  ^*:	[GetChX;  Val:= Val * Factor];
	  ^/:	[GetChX;  Val:= Val / Factor];
	  ^\:	[GetChX;  Val:= Rem(Val / Factor)]	\modulo
	other	quit;
return Val;
];	\Term



proc	AlgExp;		\Returns Term + Term, etc.
int	Val;
[Val:= Term;
loop	case Char of
	  ^+:	[GetChX;  Val:= Val + Term];
	  ^-:	[GetChX;  Val:= Val - Term]
	other	quit;
return Val;
];	\AlgExp



func	LogExp;		\Returns a boolean, e.g: AlgExp < AlgExp
int	Val;
[if Char = ^~ then	\unary 'NOT' operator
	[GetChX;
	Val:= not LogExp;
	]
else	[Val:= AlgExp;
	case Char of
	 ^=:	[GetChX;  Val:= Val = AlgExp];
	 ^#:	[GetChX;  Val:= Val # AlgExp];
	 ^>:	[GetChX;
		if Char#^= then Val:= Val > AlgExp
		else [GetChX;  Val:= Val >= AlgExp];
		];
	 ^<:	[GetChX;
		if Char#^= then Val:= Val < AlgExp
		else [GetChX;  Val:= Val <= AlgExp];
		]
	other	[];
	];
return Val;
];	\LogExp



func	BoolTerm;	\Returns LogExp & LogExp
int	Val;
[Val:= LogExp;
loop	[if Char = ^& then
		[GetChX;  Val:= Val & LogExp]
	else quit;
	];
return Val;
];	\BoolTerm



[	\Expression
SkipTab;		\must not have a tab or space
Val:= BoolTerm;
loop	[if Char = ^! then
		[GetChX;  Val:= Val ! BoolTerm]
	else quit;
	];
return Val;
];	\Expression



[	\Eval
Trap($FFFE);		\ignore divide by 0 error
GetCh;
Global:= Expression;
Trap($FFFF);
if GetErr # 0 then SetStat("div by 0", true);
if Char # ^) then SetStat("express?", true);
return EPtr;
];	\Eval

\=============================== MACRO COMMANDS ================================

proc	Cmd2Text;	\Copy command line into edit Buffer
int	I;
[for I:= 0 to DInx(DN)-1 do
	Insert(DBuf(DN, I));
];	\Cmd2Text



proc	Text2Cmd;	\Copy text into command line
int	C, D;
[C:= C2Inx;
D:= 0;
loop	[DBuf(DN, D):= Buffer(C);
	C:= C+1;
	D:= D+1;
	if D >= 2 then if DBuf(DN,D-1)=Esc & DBuf(DN,D-2)=Esc then
		[DInx(DN):= D;  quit];
	if C>=BufSize ! D>=DBufSize then
		[SetStat("esc esc?", true);
		quit];
	];
ShowCmdLine;
];	\Text2Cmd



proc	SelCmd;		\Select a command line [0-9]
int	N;
[StatHeight:= 3;  ShowStatus;	\get the number of the command line
CurL:= Height-1;  CurH:= 15;	\move cursor to status line
Cursor(CurH+1, CurL);		\but first blank 2-digit command-line number
Attrib(StatAtt);
ChOut(TV, Sp);
N:= KeyIn($0C);
if N>=^0 & N<=^9 then [DN:= N-^0;  ShowCmdLine]
else SetStat("number?", true);
Attrib($07);
CycleFlag:= false;
];	\SelCmd



proc	PreIncCmdNo;	\Increment the command-line number before entering line
[DN:= if DN>=DNSize-1 ! DN<10 then 10 else DN+1;
ShowCmdLine;
];	\PreIncCmdNo



proc	IncCmdNo;	\Increment the command-line number
[DN:= if DN >= DNSize-1 then 0 else DN+1;
CycleFlag:= true;	\re-enable command line cycling after a ^P <digit>
ShowCmdLine;
];	\IncCmdNo



proc	DecCmdNo;	\Decrement the command-line number
[DN:= if DN <= 0 then DNSize-1 else DN-1;
CycleFlag:= true;	\re-enable command line cycling after a ^P <digit>
ShowCmdLine;
];	\DecCmdNo

\================================ GET COMMANDS =================================

proc	BadCmd;		\Handle unimplemented commands
[SetStat("illegal?", true);
];	\BadCmd



proc	DoExtCmd(Ch);	\Execute an extended key command
int	Ch;		\ or insert the text character (Ch)
int	S2F;		\save the current search direction
[Ch:= -Ch;		\extended keystrokes were negated
case Ch of
RTAKEY: FwdChar;
ENDKEY: FwdAll;
HOMKEY: BackAll;
DelKey: DelForward;
PDNKEY: FwdPage;
PUPKEY: BackPage;
LFAKEY: BackChar;
DNAKEY: [DownLine;return];
UPAKEY: [UpLine;return];
SHFTAB: Indent;
INSKEY: Params(OverFlag):= not Params(OverFlag);
AltD:   BoxMode:= if BoxMode>=4 then 0 else BoxMode+1;
AltR:   CutOn;
AltF:   if CutFlag then [CutText;  BlockFill];
AltX:	DelRest;
AltB:	BackWord;
AltC:	DelWord;	\DelFwdWord;
AltN:	FwdWord;
AltO:	Oops;
FN1KEY: [DecCmdNo;  SetStat("ok", true)];
FN2KEY: [IncCmdNo;  SetStat("ok", true)];
FN3KEY: [Params(WrapFlag):= not Params(WrapFlag);
	if Params(WrapFlag) then
		if CurH >= Params(BellCol) then WrapLine;
	ShowRuler;
	];
FN4KEY: [NextBuf;  BackAll];
CtrlPgDn: Shuffle(1);
CtrlEnd: Shuffle(Height-StatHeight);
FN5KEY: CopyX;
ShiftFN5: BlockIns;
FN6KEY: DoMargin;
FN7KEY: ShowLineNo;
FN8KEY: XInx:= 0;
FN9KEY: DropMark;
FN10KEY: LastMark;
ShiftFN1: [Help(TV);  ShowCmdLine];
ShiftFN2: [Ch:= DoASCII;  if Ch>0 then Insert(Ch);  ShowCmdLine];
AltY:	[BlockIns;  XInx:= 0];
ShiftFN3:StatusOnOff;
ShiftFN4:XHairOnOff;
ShiftFN7:ScrollOnOff;
ShiftFN8:[ChOut(0, FF);  Clear;  TV:= TV+$100;  if TV>$206 then TV:= 6;
	  SetVideo;  ShowCmdLine];
CtrlFN9:		\RPi has no CtrlFN9 and no ShiftFN9 !
	[S2F:= Params(S2Flag);  Params(S2Flag):= true;
	Search(0);  Params(S2Flag):= S2F;
	];
CtrlFN10:		\RPi has no CtrlFN10 and no ShiftFN10
	[S2F:= Params(S2Flag);  Params(S2Flag):= false;
	Search(0);  Params(S2Flag):= S2F;
	]
other Insert(-Ch);	\use original negative value to distinguish from ASCII
WorkCol:= -1;		\flag new column
];	\DoExtCmd



proc	DoImmCmd(Ch);	\Execute an immediate command (control character)
int	Ch;		\ or insert the text character (Ch)
[\For speed and clarity, handle extended keys separately
if Ch < 0 then DoExtCmd(Ch)
else	[case Ch of
	 CtrlA:	FwdChar;
	 CtrlB:	OldBackWord;
	 CtrlC:	OldDelWord;
	 CtrlD:	FwdAll;
	 CtrlE:	BackAll;
	 CtrlF:	if CutFlag then CutText else XOff;
	 CtrlG:	SetStat("use J?", true); \(recursion can overflow stack)
	 CtrlH: Delete;
	 \CtrlI is a Tab
	 CtrlJ:	FwdPage;
	 CtrlK:	BackPage;
	 \CtrlL is a FF
	 \CtrlM is a CR
	 CtrlN:	OldFwdWord;
	 CtrlO:	Oops;
	 CtrlP:	SelCmd;
	 CtrlQ:	BackChar;
	 CtrlR:	XOn;
	 CtrlS:	FwdLine;
	 CtrlT:	Params(ShowFlag):= not Params(ShowFlag);
	 CtrlU: Indent;
	 CtrlV:	Cmd2Text;
	 CtrlW:	BackLine;
	 CtrlX:	DelLine;
	 CtrlY:	[CopyX;  XInx:= 0];
	 CtrlZ:	FlipCase;
	 \\Ctrl@ $00: Insert(Sp);	\because of the (un)Wyse terminal
	 \Ctrl[ is Esc
	 \Ctrl\ $1C: BadCmd;
	 \Ctrl]\$1D: [Params(S1Flag):= not Params(S1Flag);  StatHeight:= 3];
	 \Ctrl^\$1E: Insert(KeyInX);	\insert next character literally
	 \Ctrl_\$1F: [Params(S2Flag):= not Params(S2Flag);  StatHeight:= 3]
	other	Insert(Ch);
	WorkCol:= -1;			\flag new column
	];
];	\DoImmCmd

\-------------------------------------------------------------------------------

func	Conditional(D);	\Handle conditionals
int	D;
int	BC, P;
[if Global then return D;	\if true just return and execute block
BC:= 1;				\skip to matching bracket
for P:= D to DInx(DN)-1 do	\scan for loop terminator
	case DBuf(DN, P) of
	  ^]:	[BC:= BC-1;  if BC<1 then return P+1];
	  ^[:	BC:= BC+1
	other	[];
SetStat("missed ]", true);
return P;
];	\Conditional



proc	DoCmdLine(DS);	\Execute a deferred command (from the command line)
int	DS;		\starting index to command line buffer
int	I, Ch, Ch2,
	OldCh,		\last command for detecting meta (^)
	D,		\working index to command line buffer
	DX,		\index pointer to next command
	Count;		\repeat count
			\Note that groups of commands can be repeated, each
			\ group is separated by an Esc


	proc	CallCmd;	\Execute a command line (recursive)
	int	Cmd, SDN;
	[Cmd:= DBuf(DN, D+1) - ^0;	\get the command-line number
	if Cmd>=0 & Cmd<=9 & Cmd#DN then \don't allow recursion (it blows up)
		[SDN:= DN;		\save current command line
		DN:= Cmd;		\execute called command line
		DoCmdLine(0);
		SetStat("ok", false);	\don't stop now
		DN:= SDN;		\resume current command line
		]
	else	SetStat("number?", true);
	DX:= D+1;			\skip the command-line number
	];	\CallCmd


	proc	Loop;		\Execute code inside a loop (recursive)
	[DX:= DoCmdLine(D+1);
	\Abort for loop if error
	if ErrFlag then
		[I:= Huge;  if not BreakFlag then SetStat("ok", false)];
	];	\Loop


[	\DoCmdLine
Count:= 0;
WorkCol:= -1;
OldCh:= 0;
D:= DS;
loop	[if D >= DInx(DN) then quit;	\loop over all chars in command line
	OldCh:= Ch;
	Ch:= DBuf(DN, D);

	if Ch>=^0 & Ch<=^9 then		\if numeric get repeat count
		[Count:= 0;
		while Ch>=^0 & Ch<=^9 do	\get repeat count
			[Count:= Count*10 + (Ch-^0);
			if Count < 0 then [SetStat("number?", true);  quit];
			D:= D+1;
			Ch:= DBuf(DN, D);
			];
		];
	if Count < 1 then Count:= 1;	\do it at least once

	if Ch = ^# then
		[Count:= Huge;		\use huge number
		D:= D+1;
		Ch:= DBuf(DN, D);
		];

	if Ch>=^a & Ch<=^z then		\convert to uppercase if no meta
		if OldCh # ^^ then Ch:= Ch-$20;

	DX:= D+1;			\assume a one-char command

	\Execute the command Count number of times:
	for I:= 1 to Count do
		[Count:= 0;
		if (Ch<$20 & Ch#Esc & Ch#CR) ! Ch=$7F then DoImmCmd(Ch)
		else
		case Ch of			\DEFERRED COMMANDS
		  ^A:	Append;
		  ^B:	DX:= GlobalCmd(D+1);	\display or insert Global
		  ^C:	CloseFile;
		  ^D:	DelChar;		\(inserting Del is awkward)
		  ^E:	XInx:= 0;		\empty scoop
		  ^F:	DX:= ReplaceStr(D+1);	\Replace string
		  ^G:	NextBuf;		\read next text buffer
		  ^H:	DX:= InsertHex(D+1);	\insert literal hex
		  ^I:	DX:= InsertStr(D+1);	\insert string
		  ^J:	DX:= DS;		\restart command line
		  ^K:	DelAll;			\kill entire buffer
		  ^L:	List;			\list on printer
		  ^M:	DX:= Margin(D+1);	\re-margin
		  ^N:	DX:= BufSearch(D+1);	\search "N" buffers
		  ^O:	DX:= OCmd(D+1);		\open or rewind files
		  ^P:	CallCmd;		\execute another command line
		  ^Q:	Quit;			\save file and return to OS
		  ^R:	DelRest;		\delete the rest of a line
		  ^S:	DX:= DirSearch(OldCh, D+1); \search for string
		  ^T:	Terminate;		\terminate & trash output file
		  ^U:	[];			\go to OS
		  ^V:	[Text2Cmd;  quit];	\text to command line
		  ^W:	WriteBuf;		\write edit Buffer to outfile
		  ^X:	CopyX;			\copy scoop (X register)
		  ^Y:	Center;			\center a line of text
		  ^Z:	Zap;			\delete back to form feed
		  ^(:   DX:= Eval(D+1);		\evaluate an expression
		  ^{:   Loop;			\start of a loop
		  ^}:	return D+1;		\end of a loop
		  ^;:   SetStat("no loop", true);\forced exit from loop
		  ^::   MoveCur(Global);	\move cursor to Global
		  ^_:	Count:= Global;		\set count to Global value
		  ^[:   if Count>1 then SetStat("[] > 1", true)
			else DX:= Conditional(D+1); \execute block conditionally
		  ^?:	[Help(TV);  ShowCmdLine];\show help messages
		  ^/:	Help(PR);		\print help messages
		  ^]:   [];			\null if executed
		  Sp,CR:[];			\inert chars (for readability)
		  ^-,^+:[];			\sometimes arguments
		  Esc:	[]			\(command separator)
		other BadCmd;

		\Break key aborts out from all levels of recursion
		if KeyHit then
			[Ch2:= ChIn(1);
			if Ch2 = $1A \Break = weird code # 0\ then
			  [OpenI(0);  BreakFlag:= true;  SetStat("break", true)]
			else InsertKey(Ch2);
			];
		if BreakFlag then quit;
		if ErrFlag then
			[for D:= DX to DInx(DN)-1 do   \scan for loop terminator
				if DBuf(DN, D) = ^} then return D+1;
			quit;
			];
		];	\for loop

	D:= DX;				\point to next command in line buffer
	];	\loop
];	\DoCmdLine

\-------------------------------------------------------------------------------

proc	EnterCmdLine;	\Enter and execute a command line
int	Ch, SDInx;
[StatHeight:= 3;
if CycleFlag then [PreIncCmdNo;  ShowStatus];
DBuf(DN, 0):= Esc;			\store and echo the escape key
DInx(DN):= 1;
LeftCol:= 0;
SDInx:= 0;

ShowCmdLine;				\show Esc and cleared command line
					\this also sets the cursor position
repeat	[Ch:= KeyIn($0C);
	case Ch of
	-ShiftFN1:[Help(TV);  Display;  ShowCmdLine];
	-ShiftFN2:
		[Ch:= DoASCII;
		if Ch > 0 then
		    if DInx(DN) < DBufSize then
			[DBuf(DN, DInx(DN)):= Ch;  DInx(DN):= DInx(DN)+1]
		    else SetStat("length?", true);	\(do not exit loop!)
		Display;
		ShowCmdLine;
		];
	CtrlH:	if DInx(DN) > 0 then DInx(DN):= DInx(DN)-1;
	CtrlX:	DInx(DN):= 0;
	CtrlO:	if SDInx = 0 then
			[SDInx:= DInx(DN);	\Oops! Undelete command line
			DInx(DN):= 0;		\move forward until esc, esc
			repeat DInx(DN):= DInx(DN)+1;
			until (DBuf(DN, DInx(DN)-1) = Esc  &
				DBuf(DN, DInx(DN)) = Esc) ! DInx(DN) >=DBufSize;
			ShowCmdLine;
			return;
			]
		else	[			\put things back (toggle Ctrl+O)
			\DInx(DN):= SDInx;
			\SDInx:= 0;
			];
	\Ctrl^\$1E:
		[Ch:= KeyIn($0C);		\literal insert
		if DInx(DN) < DBufSize then
			[DBuf(DN, DInx(DN)):= Ch;  DInx(DN):= DInx(DN)+1]
		else SetStat("length?", true);	\(do not exit loop!)
		]
	other	[if DInx(DN) < DBufSize then
			[DBuf(DN, DInx(DN)):= Ch;  DInx(DN):= DInx(DN)+1]
		else SetStat("length?", true);	\(do not exit loop!)
		];
	ShowCmdLine;
	ShowStatus;
	];
until	(Ch=Esc & DBuf(DN, DInx(DN)-2)=Esc) ! DInx(DN)=0;      \Esc Esc or empty

DoCmdLine(0);
];	\EnterCmdLine

\-------------------------------------------------------------------------------

proc	CmdInit;	\Initialize the command buffers
int	I;
[for DN:= 0 to DNSize-1 do		\init empty (esc, esc is for Oops)
	[DInx(DN):= 0;
	for I:= 0 to DBufSize-1 do DBuf(DN,I):= Esc;
	];
DN:= 10;
LoadConfig;
];	\CmdInit



proc	Initialize;	\First-time (cold-start) initialization
int	I, J;
[OpenFiles;				\open files before doing anything else
CInx:= 0;				\empty main buffer
C2Inx:= BufSize;
SavePtrs;
XInx:= 0;				\empty X buffer (register)
ScoopOn:= false;
CutFlag:= false;

SetStat("ok", false);			\init status display and ErrFlag

CurH:= 0;  CurL:= Height - 6;

RealCol:= 0;

ChOut(0, FF);
ShowCmdLine;

Cursor(0, Height-2);			\display starting message while waiting
Text(TV, "             -- XED 3.4 ");
Text(TV, " --               For help press: Shift+F1");

\Set up ruler display:
for I:= 0 to RulerSize-1 do TickMarks(I):= ^-;
for I:= 5 to RulerSize-1 do [TickMarks(I):= ^+;   I:= I+9];
J:= ^0;
for I:= 0 to RulerSize-1 do
	[TickMarks(I):= J;
	J:= J+1;
	if J = ^: then J:= ^A;
	I:= I+9];

Append;
SavePtrs;
BackAll;
];	\Initialize

\-------------------------------------------------------------------------------

int	II;
[	\Main
InFile:= Reserve(NameSize);		\reserve all global arrays here
OutFile:= Reserve(NameSize);
OutTemp:= Reserve(NameSize);
OutBack:= Reserve(NameSize);
DInx:= Reserve(DNSize*IntSize);
DBuf:= Reserve(DNSize*IntSize);
for II:= 0 to DNSize-1 do DBuf(II):= Reserve(DBufSize*IntSize);
HexStr:= Reserve(12);
TickMarks:= Reserve(RulerSize);
Marker:= Reserve(MarkMax*IntSize);

Params:=
       [0,		\LeftMargin
	72,		\RightMargin
	72,		\BellCol
	0,		\ParaIndent
	false,		\S1Flag - ignore case
	false,		\S2Flag - search backwards
	false,		\ShowFlag
	false,		\WrapFlag
	false];		\OverFlag
StatHeight:=3;		\status lines height
BoxChar:=
       [[0],		\box character set
	\                                         
	[196, 192, 193, 217, 195, 197, 180, 218, 194, 191, 179],
	[205, 212, 207, 190, 198, 216, 181, 213, 209, 184, 179],
	[196, 211, 208, 189, 199, 215, 182, 214, 210, 183, 186],
	[205, 200, 202, 188, 204, 206, 185, 201, 203, 187, 186] ];

Global:= 0;
WorkCol:= -1;
MarkPtr:= 0;
for II:= 0 to MarkMax-1 do Marker(II):= 0;

BoxMode:= 0;		\start with box mode off
CycleFlag:= true;

TV:= $6;
SetVideo;
OutHand:=0; InHand:=0;
HaveCR:= false;

CmdInit;
ScrollMode:= true;
StatusMode:= true;
XHairMode:= false;

Scoop:= Reserve(ScoopSize);
BufSize:= 999_999;			\status line only has space for 6 digits
Buffer:= Reserve(BufSize+32);		\kluge to fix bugs in DoMargin

Initialize;

TrapC(true);
Modified:= false;

loop	[ModDel:= C2Inx-CInx;
	Display;			\display the screen
	Char:= KeyInX;			\wait for keystroke if not already hit
	case of				\init status display
	  ScoopOn: [SetStat("scoop", false); StatHeight:= 3];
	  CutFlag: [SetStat("cut", false); StatHeight:= 3]
	other SetStat("ok", false);
	BreakFlag:= false;		\no Break key struck at this point

	if Char = Esc then EnterCmdLine		\enter & execute a command line
	else if Char = CtrlG then DoCmdLine(0)	\(not allowed in deferred mode)
	else DoImmCmd(Char);		\do immediate command or insert char

	if C2Inx-CInx # ModDel then Modified:= true;
	];
];	\Main
